#include "arte/arte.hh"
#include "grover/higz.h"
#include "grover/grover.hh" // - class defs, f.prototypes and global variables

//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// ------------------------------------------------------------------------
// 00-10-01  M.Sang   added GroverEvent(void) as default
//



int GroverEvent(GroverCuts* usercuts = 0) {

  /////////////////////////////////////////////////////////////////
  //
  // Routine to reconstruct all primary vertices in an event
  // and...
  //
  // Under development. This is the routine that users will usually
  // call. When secondary (J/psi+anything and/or anything) 
  // finding is stable this will go here as well.
  //
  // M.Sang 05 June 00
  //
  /////////////////////////////////////////////////////////////////

  
  // Counter and error flag
  static int  eventNumber = 0;
  eventNumber++;
  bool firstcall = (eventNumber == 1);
  int error = 0;



  // Non-default cuts?
  static GroverCuts cuts;
  if (firstcall && usercuts != 0) cuts = *usercuts; //(Memberwise copy)



  // Printout?
  static bool verbose = (cuts.Verbose == 1);



  // Do some evaluation - efficiency estimation, etc?
  static bool verona = (cuts.GPEActivate == 1);
  if (verbose && verona && firstcall) {
    cout << "*** GroverEvent - Using user-defined cuts" << endl;
  }
  else if (verbose && !verona && firstcall)
    cout << "*** GroverEvent - Using default cuts" << endl;



  // Which package?
  static bool usingGrover = false, usingVT = false;
  if (cuts.VertexingPackage == 1) {
    usingGrover = true;  usingVT = false;
  }
  else if (cuts.VertexingPackage == 2) {
    usingGrover = true; usingVT = true; // Grover needed for wire alignment
  }
  else {
    cout << "*** GroverEvent: ERROR - unknown vertexing package ID = "
	 << cuts.VertexingPackage << endl;
    return -1;
  }



  // Grover objects
  static GroverParticleList Particles;
  static GroverWireList     Wires;
  static GroverVertexList   rPrimaries, mcPrimaries;
  static GroverVertexList   rSecondaries;
  if (!firstcall) {
    Particles.clear(); Wires.clear(); 
    rPrimaries.clear(); mcPrimaries.clear(); rSecondaries.clear();
  }



  // Refit RSEGs with MC truth momentum (smeared or not) if requested
  if (cuts.CheatMomentum == 1 && ArteTable<MCIN>::booked()) {
    if (verbose) 
      cout << "*** GroverEvent: Refitting all VDS segments with MC momenta" << endl;
    int err = GroverRefitRSEG(&cuts);
    if (err != 0) 
      cout << "*** GroverEvent: WARNING - GroverRefitRSEG returned error" <<endl;
  }



  // Fill particle list from RSEG or RTRA or Rover default (changes with time)
  switch(cuts.RPTrackChoice) {
  case 1: {
    GroverFillLikeRover(&Particles);
    break;
  }
  case 2: {
    GroverFillFromRSEG(&Particles);
    break;
  }
  case 3: {
    GroverFillFromRTRA(&Particles);
    break;
  }
  default: {
    GroverFillLikeRover(&Particles);
    if (firstcall) cout << "*** GroverEvent: WARNING - unknown track selection"
			<< " -> using default! (no more warnings)" << endl;
    break;
  }
  }
  if (Particles.empty()) {
    if (verbose) {
      cout << "*** GroverEvent: WARNING - no ";
      switch(cuts.RPTrackChoice) {
      case 1: {
	cout << "particles found using standard selection" << endl;
	break;
      }
      case 2: {
	cout << "VDS segments in RSEG" << endl;
	break;
      }
      case 3: {
	cout << "tracks in RTRA" << endl;
	break;
      }
      default: {
	cout << "particles found using standard selection" << endl;
	break;
      }
      }
      error -= 1;
    }
  }



  // Move Particles to VDS local coordinates
  for (GroverParticleList::iterator ipar = Particles.begin();
       ipar != Particles.end(); ++ipar) ipar->GlobalToLocal();



  // Fill wires and move to VDS local coordinates
  GroverFillFromGTAR(&Wires);
 if (Wires.empty()) {
    if (verbose) 
      cout << "*** GroverEvent: WARNING - no active wires in GTAR" << endl;
    error -= 2;
  }
  for (GroverWireList::iterator iwire = Wires.begin();
       iwire != Wires.end(); ++iwire) {
    iwire->GlobalToLocal();
  }
  



  // Call primary finder(s)
  int flag = GroverPrimaries(&Particles, &Wires, &cuts, &rPrimaries);
  if (flag != 0) {
    if (verbose) 
      cout << "*** GroverEvent: WARNING - GroverPrimaries error = " 
	   << flag << endl;
    error -= 4;;
  }
  if (usingVT) {
    rPrimaries.clear();
    cout << "*** GroverEvent: ERROR - VT disabled" << endl;
    //flag = GroverVTPrimaries(&Particles, &Wires, &cuts, &rPrimaries);
    if (flag != 0) {
      if (verbose)
	cout << "*** GroverEvent: WARNING - GroverVTPrimaries error = " 
	     << flag << endl;
      error -= 8;;
    }
  }


  // Get MC primary vertices if required
  GroverVertexList* mcverts = 0;
  if (verona && ArteTable<MCIN>::booked()) {
    flag = GroverFillFromMVER(&mcPrimaries, 1);
    if (flag != 0) {
      if (verbose) 
	cout << "*** GroverEvent: WARNING - couldn't get "
	        "MC primary vertices" << endl;
      error -= 16;
    }
    else mcverts = &mcPrimaries;
  }



  // Calculate primary efficiencies, etc...
  if (verona) {
    flag = GroverPrimEval(&rPrimaries,mcverts,&cuts);
    if (flag != 0) {
      if (verbose) 
	cout << "*** GroverEvent: WARNING - GroverPrimEval failed" << endl;
      error -= 32;
    }
  }
  
  if (verbose) cout << "*** GroverEvent: finding 2ndaries" <<endl;

  // Find Secondaries
  if (true) {
    flag = GroverSecondaries(&Particles, &Wires, &cuts, 
			     &rPrimaries, &rSecondaries);
    if (flag != 0) {
      if (verbose) 
	cout << "*** GroverEvent: WARNING - GroverSecondaries failed" << endl;
      error -= 64;
    }
  }

  if (verbose) cout << "*** GroverEvent: found 2ndaries" << endl;

  // Clear RVER 
  ArteTable<RVER>::clear();



  // Fill RVER with primaries
  flag = GroverAddToRVER(&rPrimaries, 1);
  if (flag != 0) {
    if (verbose) 
      cout << "*** GroverEvent: WARNING - couldn't write "
	      "primary vertices into RVER" 
	   << endl;
    error -= 128;
  }


  // Fill RVER with secondaries
  if (true) {
    flag = GroverAddToRVER(&rSecondaries, 2);
    if (flag != 0) {
      if (verbose) 
	cout << "*** GroverEvent: WARNING - couldn't write "
	  "secondary vertices into RVER" 
	     << endl;
      error -= 256;
    }
  }

  return error;
}



/////////////////////////////////////////////
int GroverEvent(void)
{ // default
  return GroverEvent(0);
}
