#include <iostream.h>
#include <iomanip.h>
#include "arte/arte.hh"
#include "TROOT.h"        //
#include "TFile.h"        // ROOT stuff
#include "TNtuple.h"      //
#include "grover/grover.hh"

extern "C" {
#include "grover/hbook.h"
}

//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// -----------------------------------------------------------------------------
//
//

TROOT    GroverRoot("GroverRoot", "Grover ROOT object");
TFile*   GroverOutputFile; 
TNtuple* RecoNtuple;
TNtuple* MCNtuple;

////////////////////////////////////////////////////////////////////////

int GroverPrimEval(GroverVertexList* rv   = 0,
		   GroverVertexList* mcv  = 0,
		   GroverCuts*       cuts = 0)
{
  ///////////////////////////////////////////////////////////////
  //
  // Grover primary vertex finder evaluation program
  //
  // Takes a list of reconstructed vertices and (optionally)
  // a list of MC vertices and calculates efficiencies, ghost
  // rates, resolutions, pulls etc. Works with PAW/HBOOK and ROOT.
  // ***Needs a GroverCuts object for steering.
  //
  // M.Sang 25/05/00
  // based on VERONA 2.2 by J.van Tilburg
  //
  ///////////////////////////////////////////////////////////////


  // bail out!
  if (rv==0 || cuts==0) return -1;


  bool empty = (rv->empty());
  static int EventNumber = 0, EmptyEventNumber = 0;
  EventNumber++; if (empty) EmptyEventNumber++;
  bool firstcall = (EventNumber==1);
  bool verbose = (cuts->Verbose == 1);
  int  update  = (cuts->GPEUpdateFreq);
  bool writing = (EventNumber%update == 0) ;
  bool MC = (mcv != 0);
  const int ASIZE = 100;
  Float_t Reco[99] = {0.0}; // if we have >99 variables in the Ntuples then 
  Float_t Simu[99] = {0.0}; // the array sizes must be increased accordingly



  // Book histograms if first event
  bool PAW  = (cuts->Paw  == 1);
  bool ROOT = (cuts->Root == 1);
  if (firstcall) {
    if (PAW) {
      // book some paw stuff
      if (PrimEvalInitPaw()!=0) 
	cout << "*** GroverPrimEval: ERROR - Root setup failed" << endl;
    }
    if (ROOT) {
      // book some root stuff
      if (PrimEvalInitRoot() != 0) 
	cout << "*** GroverPrimEval: ERROR - Root setup failed" << endl;
    }
  }



  // Fill arrays with positions, errors, etc., of found (real) 
  // and findable (mc) vertices
  float  x_mc[ASIZE]={0.0},  y_mc[ASIZE]={0.0},  z_mc[ASIZE]={0.0};
  float   x_r[ASIZE]={0.0},   y_r[ASIZE]={0.0},   z_r[ASIZE]={0.0};
  float  dx_r[ASIZE]={0.0},  dy_r[ASIZE]={0.0},  dz_r[ASIZE]={0.0};
  int    nt_r[ASIZE]={0};
  float chi2 = 0.0, chi2ptrk = 0.0;
  int   nhitmin = cuts->GPEMinHitsInTrack;
  int   ntmin   = cuts->GPEMinTracksInVertex;
  float maxchi2 = cuts->GPEMaxChi2;
  float maxchi2ptrk = cuts->GPEMaxChi2PerTrack;
  // ===== MC ======
  int   nmcv = 0;
  if (MC) {
    for (GroverVertexList::iterator imcv = mcv->begin();
	 imcv != mcv->end(); ++imcv) {
      int nmctr = 0;
      for (GroverParticleList::iterator ipar = (imcv->particles).begin();
	   ipar != (imcv->particles).end(); ++ipar) {
	if (ipar->Nhit() >= nhitmin) nmctr++; // #hits = # of reco. VDS mimps
      }
      if (nmctr >= ntmin) {
	x_mc[nmcv] = imcv->X();
	y_mc[nmcv] = imcv->Y();
	z_mc[nmcv] = imcv->Z();
	nmcv++;
      }
    }
  }
  // ====== REAL ======
  int nrv = 0;
  for (GroverVertexList::iterator irv = rv->begin();
       irv != rv->end(); ++irv) {
    int nrtr = 0;
    for (GroverParticleList::iterator ipar = (irv->particles).begin();
	 ipar != (irv->particles).end(); ++ipar)
      if (ipar->Nhit() >= nhitmin) nrtr++;
    chi2     = irv->Chi2();
    chi2ptrk = (nrtr>0) ? (chi2/static_cast<float>(nrtr)) : 1.0e6;
    if (nrtr     >= ntmin      &&
	chi2     <= maxchi2    &&
	chi2ptrk <= maxchi2ptrk
	) {
      nt_r[nrv] = nrtr; 
      x_r[nrv]  = irv->X();  
      dx_r[nrv] = (irv->Cov(0)>0.0) ? (sqrt(irv->Cov(0))) : 0.0;
      y_r[nrv]  = irv->Y();  
      dy_r[nrv] = (irv->Cov(2)>0.0) ? (sqrt(irv->Cov(2))) : 0.0;
      z_r[nrv]  = irv->Z();
      dz_r[nrv] = (irv->Cov(5)>0.0) ? (sqrt(irv->Cov(5))) : 0.0;
      nrv++;
    }
  }



  // Cumulative counters: "MC" means findable MC, "R" means reconstructed
  static int MCTotal = 0, RTotal  = 0;
  MCTotal += nmcv;
  RTotal  += nrv;
  static int MCMissed  = 0;
  static int RGood     = 0;
  static int RClone    = 0;
  static int RMerged   = 0;
  static int RMixed    = 0;
  static int RGhost    = 0;
  static int RUnknown  = 0;



  // Box sizes for efficiency/ghost rate estimation
  float Dxymax = cuts->GPETransverseWindow;
  float Dzmax  = cuts->GPELongitudinalWindow;



  // Associations between MC and real vertices
  int      NMCAssocTo[ASIZE] = {0};
  int   BestMCAssocTo[ASIZE] = {-1};
  float    CDToBestMC[ASIZE] = {1e6};
  int       NRAssocTo[ASIZE] = {0};
  int    BestRAssocTo[ASIZE] = {-1};
  float     CDToBestR[ASIZE] = {1e6};
  float ChiDistCut = cuts->GPEChiWindow;
  if (MC) {
    for (int irv=0; irv<nrv; irv++) {
      for (int imcv=0; imcv<nmcv; imcv++) {

	// Chi2 cut
	//float Dx = fabs((x_r[irv]-x_mc[imcv]) / dx_r[irv]);
	//float Dy = fabs((y_r[irv]-y_mc[imcv]) / dy_r[irv]); 
	//float Dz = fabs((z_r[irv]-z_mc[imcv]) / dz_r[irv]);
	//float D  = sqrt((Dx*Dx)+(Dy*Dy)+(Dz*Dz));
	//if (D < ChiDistCut) { 

	// Spatial cuts
	float Dx = fabs((x_r[irv]-x_mc[imcv]));
	float Dy = fabs((y_r[irv]-y_mc[imcv])); 
	float Dz = fabs((z_r[irv]-z_mc[imcv]));
	// D is only sensible if one target station is used.
	// If two stations are in use, all three dimensions must be
	// used.
	float D  = sqrt((Dx*Dx)+(Dy*Dy)); 
	if (Dx<Dxymax && Dy<Dxymax && Dz<Dzmax) {

	  (NMCAssocTo[irv])++;  (NRAssocTo[imcv])++;
	  if (D < CDToBestR[imcv]) {
	    BestRAssocTo[imcv] = irv;
	    CDToBestR[imcv] = D;
	  }
	  if (D < CDToBestMC[irv]) {
	    BestMCAssocTo[irv] = imcv;
	    CDToBestMC[irv] = D;
	  }
	}
      }
    }
  }


  
  // Classification
  int RecoType[ASIZE] = {0};
  int MCType[ASIZE]   = {0};
  if (MC) {
    // Classify the MC vertices
    for (int imcv=0; imcv<nmcv; imcv++) {
      if (NRAssocTo[imcv] == 0) {
	MCType[imcv] = 1;
	MCMissed++;
      }
    }
    // Classify the reconstructed vertices
    for (int irv=0; irv<nrv; irv++) {
      if (NMCAssocTo[irv] == 0) {
	RGhost++; 
	RecoType[irv] = 1;
      }
      else if (NMCAssocTo[irv] == 1 && 
	       BestRAssocTo[BestMCAssocTo[irv]] == irv) {
	RecoType[irv] = 2;
	RGood++;
      }
      else if (NMCAssocTo[irv]==1 &&
	       BestRAssocTo[BestMCAssocTo[irv]] != irv) {
	RecoType[irv] = 3;
	RClone++;
      }
      else if (NMCAssocTo[irv] >1) {
	RecoType[irv] = 4;
	RMixed++;
      }
      else {
	RecoType[irv] = -1;
	RUnknown++;
      }
    }
  }

  

  // Some arithmetic
  float fMCMissed  = static_cast<float>(MCMissed);
  float fMCTotal   = static_cast<float>(MCTotal);
  float fRGhost    = static_cast<float>(RGhost);
  float fRGood     = static_cast<float>(RGood);
  float fRClone    = static_cast<float>(RClone);
  float fRMixed    = static_cast<float>(RMixed);
  float fRUnknown  = static_cast<float>(RUnknown);
  float fRTotal    = static_cast<float>(RTotal);
  float Efficiency = (fMCTotal>0.0) ? 100.0*(1.0 - fMCMissed/fMCTotal) : 0.0;
  float GhostRate  = (fRTotal>0.0)  ? 100.0*(fRGhost/fRTotal)          : 0.0;
  float GoodRate   = (fRTotal>0.0)  ? 100.0*(fRGood/fRTotal)           : 0.0;
  float CloneRate  = (fRTotal>0.0)  ? 100.0*(fRClone/fRTotal)          : 0.0;
  float MixedRate  = (fRTotal>0.0)  ? 100.0*(fRMixed/fRTotal)          : 0.0;
  float UnknownRate= (fRTotal>0.0)  ? 100.0*(fRUnknown/fRTotal)        : 0.0;



  // Write out summary
  if (writing) {
    cout << "*******************"
      " GroverPrimEval: SUMMARY "
      "***************************" 
	 << endl << endl;
    cout << " Total Number of events                       " 
	 << EventNumber << endl;
    cout << " Number of events with >= 1 reco vertex       "
	 << EventNumber-EmptyEventNumber << endl;
    cout << " Number of reconstructed vertices             "
	 << RTotal << endl;
    if (MC) {
      cout << " Number of MC primary vertices                "
	   << MCTotal << endl;
      cout << " Efficiency (%)                               "
	   << Efficiency << endl;
      cout << "---- Reconstructed vertices by type ----      "
	   << endl;
      cout << " Unambiguous (%)                              "
	   << GoodRate << endl;
      cout << " Ghosts (%)                                   "
	   << GhostRate << endl;
      cout << " Clones (%)                                   "
	   << CloneRate << endl;
      cout << " Mixed/Merged (%)                             "
	   << MixedRate << endl;
      cout << " Unclassified (%)                             "
	   << UnknownRate << endl;
    }
    cout << endl
	 << "*****************************************"
            "******************************"
	 << endl;
  }



  // Fill Ntuples and/or Histos for reconstructed vertices
  if (PAW) {
    hf1(1600, static_cast<float>(nrv), 1.0);
  }
  for (int irv = 0; irv<nrv; irv++) {
    if (PAW) {
      hf1(1121, x_r[irv], 1.0);
      hf1(1122, y_r[irv], 1.0);
      hf1(1123, z_r[irv], 1.0);
      hf1(1610, nt_r[irv], 1.0);
      if (MC && BestMCAssocTo[irv]>=0) {
	hf1(1101, x_r[irv]-x_mc[BestMCAssocTo[irv]], 1.0);
	hf1(1102, y_r[irv]-y_mc[BestMCAssocTo[irv]], 1.0);
	hf1(1103, z_r[irv]-z_mc[BestMCAssocTo[irv]], 1.0);
	if (dx_r[irv]>0.0) 
	  hf1(1111, (x_r[irv]-x_mc[BestMCAssocTo[irv]])/dx_r[irv], 1.0);
	if (dy_r[irv]>0.0) 
	  hf1(1111, (y_r[irv]-y_mc[BestMCAssocTo[irv]])/dy_r[irv], 1.0);
	if (dz_r[irv]>0.0) 
	  hf1(1111, (z_r[irv]-z_mc[BestMCAssocTo[irv]])/dz_r[irv], 1.0);
      }
    }
    if (ROOT) {
      Reco[0] = static_cast<Float_t>((*rv)[irv].Wire());
      Reco[1] = static_cast<Float_t>((*rv)[irv].particles.size());
      Reco[2] = static_cast<Float_t>(x_r[irv]);
      Reco[3] = static_cast<Float_t>(y_r[irv]);
      Reco[4] = static_cast<Float_t>(z_r[irv]);
      Reco[5] = static_cast<Float_t>(dx_r[irv]);
      Reco[6] = static_cast<Float_t>(dy_r[irv]);
      Reco[7] = static_cast<Float_t>(dz_r[irv]);
      Reco[8] = (MC && (BestMCAssocTo[irv] >= 0)) ?
	static_cast<Float_t>(x_r[irv]-x_mc[BestMCAssocTo[irv]]) : -999.0;
      Reco[9] = (MC && (BestMCAssocTo[irv] >= 0)) ?
	static_cast<Float_t>(y_r[irv]-y_mc[BestMCAssocTo[irv]]) : -999.0;
      Reco[10]= (MC && (BestMCAssocTo[irv] >= 0)) ?
	static_cast<Float_t>(z_r[irv]-z_mc[BestMCAssocTo[irv]]) : -999.0;
      Reco[11]= static_cast<Float_t>((*rv)[irv].Chi2());
      Reco[12]= static_cast<Float_t>(RecoType[irv]);
      RecoNtuple->Fill(Reco);
    }
  }
  for (int imcv=0; imcv<nmcv; imcv++) {
    if (PAW) {
    }
    if (ROOT) {
      Simu[0] = static_cast<Float_t>((*mcv)[imcv].Wire());
      Simu[1] = static_cast<Float_t>((*mcv)[imcv].particles.size());
      Simu[2] = static_cast<Float_t>(x_mc[imcv]);
      Simu[3] = static_cast<Float_t>(y_mc[imcv]);
      Simu[4] = static_cast<Float_t>(z_mc[imcv]);
      Simu[5] = static_cast<Float_t>(MCType[imcv]);
      MCNtuple->Fill(Simu);
    }
  }


  return 0;
}


////////////////////////////////////////////////////////////////////////



int PrimEvalInitPaw(void)
{
  hbook1( 1121, "reco vtx x",             200,  -0.20,   0.80);
  hbook1( 1122, "reco vtx y",             200,  -0.55,   0.45);
  hbook1( 1123, "reco vtx z",             240,  -6.00,   0.00);
  hbook1( 1101, "x_rec - x_mc",            50,  -0.10,   0.10);
  hbook1( 1102, "y_rec - y_mc",            50,  -0.10,   0.10);
  hbook1( 1103, "z_rec - z_mc",            50,  -0.50,   0.50);
  hbook1( 1111, "x_rec - x_mc pull",       30, -10.00,  10.00);
  hbook1( 1112, "y_rec - y_mc pull",       30, -10.00,  10.00);
  hbook1( 1113, "z_rec - z_mc pull",       30, -10.00,  10.00);
  hbook1( 1600, "Reco vtices per event",   11,  -0.50,  10.50);
  hbook1( 1610, "Num of tracks in vertex", 26,  -0.50,  25.50);
  return 0;
}

////////////////////////////////////////////////////////////////////////


int PrimEvalInitRoot(void)
{
  // Open file
  GroverOutputFile = 
    new TFile("Vertices.root", "RECREATE", "Primary vertex ntuples", 1);
  if (GroverOutputFile->IsOpen()) {
    cout << "*** GroverPrimEval: Opened ROOT file" << endl;
  }
  else 
    cout << "*** GroverPrimeEval: ERROR - could not open ROOT file" << endl;


  // Book ntuples
  RecoNtuple = new TNtuple("RecoNtuple", "Reconstructed primary vertices", 
			   "wire:ntrks:x:y:z:dx:dy:dz:"
			   "pullx:pully:pullz:chi2:type");
  MCNtuple   = new TNtuple("MCNtuple", "Reconstructable MC primary vertices", 
			    "wire:ntrks:x:y:z:type");
  cout << "*** GroverPrimEval: Finished initialisation" << endl;
  return 0;
}


////////////////////////////////////////////////////////////////////////
