#include <iostream>
#include <algorithm>
#include "arte/arte.hh"
#include "arte/RSEG_cmp.hh"
#include "grover/grover.hh"
#include "grover/higz.h"

//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// -----------------------------------------------------------------------------
//
//


int GroverFindMCJpsi(GroverVertexList*);

int GroverSecondaries(GroverParticleList* allParticles,
		      GroverWireList*     allWires,
		      GroverCuts*         cuts,
		      GroverVertexList*   allPrimaries,
		      GroverVertexList*   allSecondaries) 
{
  /////////////////////////////////////////////////////////////////////
  //
  //                GroverSecondaries
  // Finds secondary vertices consisting of a) J/psi + zero or more trks
  //                                        b) anything
  //                                        c) lepton + anything (not yet)
  // Author M.Sang
  //
  /////////////////////////////////////////////////////////////////////

  static const float m_e = 0.000511;


  static int EventNumber = 0;
  static bool firstcall = true;
  EventNumber++;
  firstcall = (EventNumber==1);

  static bool verbose = (cuts->Verbose == 1);


  // Initialisation
  if (firstcall || verbose)
    cout << "********* Grover v."
	 << GroverVersion 
	 << ": Finding Secondaries using Grover ******" << endl;
  


  //******** 1) Find Secondaries consisting of J/psi + anything charged
  static GroverVertexList jpsi_list;
  if (cuts->FindJpsiPlusX == 1) {
    // Find J/psi candidate in RTRA, get corresponding RSEGs and momenta
    // remove the RSEGs from the particle list and turn J/psi tracks
    // into a vertex - push onto list and call GroverJpsiSec
    if (!jpsi_list.empty()) jpsi_list.clear();
    
    // find J/psi (NOTE!! assumes only one J/psi in RTRA!)
    bool foundpsi = false;
    ArtePointer<RTRA> jpsi;
    for (ArteTable<RTRA>::iterator irtra = ArteTable<RTRA>::begin();
	 irtra != ArteTable<RTRA>::end(); ++irtra) {
      if (irtra->fit == 2 && !foundpsi) {
	jpsi=irtra;
	if (verbose) cout << "*** GroverSecondaries: FOUND PSI (rtra " 
			  << jpsi << ")" << endl;
	foundpsi = true;
      }
    }  
    ArtePointer<RSEG> lep1=0, lep2=0;
    bool onefound=false, twofound=false;
    float p1=0.0, p2=0.0;
    if (foundpsi) {

      //if (verbose) cout << "found the jpsi" << endl;
      ArtePointer<RVER> jpver =  jpsi->ever;
      ArteRelationIterator<RREL> rrel1p = jpver->rrel.begin();
      ArteRelationIterator<RREL> rrel2p = rrel1p; rrel2p++;
      ArtePointer<RREL> rrel1 = *rrel1p;
      ArtePointer<RREL> rrel2 = *rrel2p;
      ArtePointer<RTRA> rtra1 = rrel1->rtra;
      ArtePointer<RTRA> rtra2 = rrel2->rtra;


      // ***** Old way - J/psi from RSEGs ******** //
      if (false) {
	p1 = rtra1->pf; p2 = rtra2->pf;
	if (verbose) cout << "*** GroverSecondaries: rver,rtra1,rtra2 = "
			  << jpver << ", " << rtra1 << ", " << rtra2 << endl;
	if (fabs(p1)>0.0) p1 = fabs(1.0/p1);
	if (fabs(p2)>0.0) p2 = fabs(1.0/p2);
	for (ArteRelationIterator<RSEG> irseg = rtra1->rseg.begin();
	     irseg != rtra1->rseg.end(); ++irseg) {
	  if (!((*irseg)->cmp & ~(Rsegc::vxd | Rsegc::bit16))) {
	    lep1     = *irseg; 
	    onefound = true;
	  }
	}
	for (ArteRelationIterator<RSEG> jrseg = rtra2->rseg.begin();
	     jrseg != rtra2->rseg.end(); ++jrseg) {
	  if (!((*jrseg)->cmp & ~(Rsegc::vxd | Rsegc::bit16))) {
	    lep2     = *jrseg; 
	    twofound = true;
	  }
	}
      }
      // ********* New way - J/psi from RTRAs *********** //
      if (rtra1!=0 && rtra2!=0)	onefound = twofound = true;
      

      if (onefound && twofound) {
	if (verbose) 
	  cout << "*** GroverSecondaries: found leptons - making vertex" 
	    //<< " from RSEGs " << lep1 << ", " << lep2 
	       << endl;
	// make a vertex out of the daughters
	// *** old way
	//GroverParticle lepton1(lep1), lepton2(lep2);
	// *** new way
	GroverParticle lepton1(rtra1), lepton2(rtra2);
	if (verbose) cout 
	  << "*** GroverSecondaries: J/psi rver,rtra1,rtra2,p1,p2 = " 
	  << jpver << " " << rtra1 << " " << rtra2
	  << rtra1->pf << " " << rtra2->pf << endl;
	// change coordinates and set momenta and masses
	lepton1.GlobalToLocal(); lepton2.GlobalToLocal();
	//lepton1.SetOMomentum(p1); lepton2.SetOMomentum(p2);
	lepton1.SetMass(m_e);    lepton2.SetMass(m_e);
	lepton1.PropagateBeginTo(-3.0); lepton2.PropagateBeginTo(-3.0);
	if (verbose) cout << "  -- LEPTON MOMENTA = " 
			  << lepton1.Momentum() << " " << lepton2.Momentum() << endl;
	GroverVertex jpcand(&lepton1); jpcand.Add(&lepton2);
	// flag the electrons in the global list
	for (GroverParticleList::iterator ipar = allParticles->begin();
	     ipar != allParticles->end(); ++ipar) {

	  // *** old way
	  if ((ipar->ParentRSEG() == lep1 || ipar->ParentRSEG() == lep2) && false) {
	    ipar->SetUsedInSecondary(1234);
	    if (verbose) cout <<"flagging segment " 
			      << ipar->ParentRSEG() << " as used" <<endl;
	  }
	  // **** new way
	  if ((ipar->ParentRTRA() == rtra1 || ipar->ParentRTRA() == rtra2) && true) {
	    ipar->SetUsedInSecondary(1234);
	    if (verbose) cout <<"flagging track " 
			      << ipar->ParentRTRA() << " as used" <<endl;
	  }
	}
	if (true) {
	  jpsi_list.push_back(jpcand); // This is the normal routine!!
	  jpcand.GlobalToLocal();
	  // See what we can do with it!
	  if (verbose) cout 
	    << "*** GroverSecondaries: Calling GroverJpsiSec" << endl;
	  int error = GroverJpsiSec(allParticles, allWires, cuts, 
				    allPrimaries, &jpsi_list);
	  if (verbose) cout << "*** GroverSecondaries: done jpsi+X" << endl;
	}
      }
    }
    //=======================================================================
    if (false) { // this is a dirty hack for the 4 track MC
      // -> find the Jpsi segments by cheating, make a vertex out of them using
      // the true momenta.
      int err2 = GroverFindMCJpsi(&jpsi_list);
      if (err2 == 0) {
	// flag the electrons in the global list
	GroverParticleList::iterator p1 = (jpsi_list.begin())->particles.begin();
	GroverParticleList::iterator temp = p1;
	GroverParticleList::iterator p2 = ++temp;
	ArtePointer<RSEG> rs1=p1->ParentRSEG(), rs2=p2->ParentRSEG();
	for (GroverParticleList::iterator ipar = allParticles->begin();
	     ipar != allParticles->begin(); ++ipar) {
	  if (ipar->ParentRSEG() == rs1 || ipar->ParentRSEG() == rs2)
	    ipar->SetUsedInSecondary(1234);
	}
	// See what we can do with it!
	if (verbose) cout << "*** GroverSecondaries: Calling GroverJpsiSec" << endl;
	int error = GroverJpsiSec(allParticles, allWires, cuts, 
				  allPrimaries, &jpsi_list);
	if (verbose) cout << "*** GroverSecondaries: done jpsi+X" << endl;
      }
    }
    //======================================================================

  }





  
  // ******** 2) Find secondaries consisting of anything else
    static GroverVertexList unbiased_list;
  if (cuts->FindUnbiased == 1) {
    if (verbose) cout << "*** GroverSecondaries: Calling GroverUnbiasedSec." << endl;
    if (!unbiased_list.empty()) unbiased_list.clear();
    int eflag = GroverUnbiasedSecondaries(allParticles, allWires, cuts, 
					  allPrimaries, &unbiased_list);  
  }
    






  // ********* 3) Find secondaries consisting of 1 lepton from 
  //              the J/psi and anything else (ie charm search)
  static GroverVertexList charm_list;
  if (cuts->FindCharm == 1) {
    cout << "GROVER CHARM NOT AVAILABLE IN OFFICIAL GROVER!!!!!!" << endl;
    // Find J/psi candidate in RTRA, get corresponding RSEGs and momenta
    // remove the RSEGs from the particle list
    if (!charm_list.empty()) charm_list.clear();
    // find J/psi (NOTE!! assumes only one J/psi in RTRA!)
    bool foundpsi = false;
    ArtePointer<RTRA> jpsi;
    for (ArteTable<RTRA>::iterator irtra = ArteTable<RTRA>::begin();
	 irtra != ArteTable<RTRA>::end(); ++irtra) {
      if (irtra->fit == 2 && !foundpsi) {
	jpsi=irtra;
	if (verbose) cout << "*** GroverSecondaries: FOUND PSI" << endl;
	foundpsi = true;
      }
    }  
    ArtePointer<RSEG> lep1=0, lep2=0;
    bool onefound=false, twofound=false;
    float p1=0.0, p2=0.0;
    if (foundpsi) {
      ArtePointer<RVER> jpver =  jpsi->ever;
      ArtePointer<RREL> rrel1 = *(   jpver->rrel.begin() );
      ArtePointer<RREL> rrel2 = *(++(jpver->rrel.begin()));
      ArtePointer<RTRA> rtra1 = rrel1->rtra;
      ArtePointer<RTRA> rtra2 = rrel2->rtra;
      p1 = rtra1->pf; p2 = rtra2->pf;
      if (verbose) cout << "*** GroverSecondaries: rver,rtra1,rtra2 = "
			<< jpver << ", " << rtra1 << ", " << rtra2 << endl;
      if (fabs(p1)>0.0) p1 = fabs(1.0/p1);
      if (fabs(p2)>0.0) p2 = fabs(1.0/p2);
      for (ArteRelationIterator<RSEG> irseg = rtra1->rseg.begin();
	   irseg != rtra1->rseg.end(); ++irseg) {
	if (!((*irseg)->cmp & ~(Rsegc::vxd | Rsegc::bit16))) {
	  lep1     = *irseg; 
	  onefound = true;
	}
      }
      for (ArteRelationIterator<RSEG> jrseg = rtra2->rseg.begin();
	   jrseg != rtra2->rseg.end(); ++jrseg) {
	if (!((*jrseg)->cmp & ~(Rsegc::vxd | Rsegc::bit16))) {
	  lep2     = *jrseg; 
	  twofound = true;
	}
      }
      if (onefound && twofound) {
	if (verbose) cout 
	  << "*** GroverSecondaries: found leptons" 
	  << " from RSEGs " << lep1 << ", " << lep2 << endl;
	// flag the leptons in the global particle list
	for (GroverParticleList::iterator ipar = allParticles->begin();
	     ipar != allParticles->end(); ++ipar) {
	  if (ipar->Type()==2) {
	    if (ipar->ParentRSEG()==lep1 || ipar->ParentRSEG()==lep2) {
	      ipar->SetUsedInSecondary(1234);
	      if (verbose) cout << "flagged a particle as a lepton" << endl;
	    }
	  }
	}
	// Now run the charm finder (hacked unbiased, but including flagged
	// leptons as seeds even if they have bad IPS)
	if (verbose) cout 
	  << "*** GroverSecondaries: Calling GroverCharm" << endl;
	int error = GroverCharm(allParticles, allWires, cuts, 
				allPrimaries, &charm_list);
      }
    }
    else cout << "No Jpsi found in RTRA	!" <<endl;
  }



    
  if (verbose) {
    int nsec = jpsi_list.size() + unbiased_list.size() + charm_list.size();
    cout << "*** GroverSecondaries: found ";
    if (nsec==1) cout << "1 secondary vertex" << endl;
    else cout << nsec << " secondary vertices" << endl;
  }
  



  // 4) Add secondary lists to allSecondaries
  // a) J/psi + anything
  if (cuts->FindJpsiPlusX ==1) {
    for (GroverVertexList::iterator iver = jpsi_list.begin();
	 iver != jpsi_list.end(); ++iver) {
      allSecondaries->push_back(*iver);
    }
  }
  // b) Anything
  if (cuts->FindUnbiased ==1) {
    for (GroverVertexList::iterator iver = unbiased_list.begin();
	 iver != unbiased_list.end(); ++iver) {
      allSecondaries->push_back(*iver);
    }
  }
  // c) Charm
  if (cuts->FindCharm ==1) {
    for (GroverVertexList::iterator iver = charm_list.begin();
	 iver != charm_list.end(); ++iver) {
      allSecondaries->push_back(*iver);
    }
  }
  
  // 4) Draw secondary vertices
  float xpm[1], ypm[1];
  if (cuts->GroverDraw == 1 &&
      (cuts->FindJpsiPlusX ==1 ||
       cuts->FindUnbiased  ==1 ||
       cuts->FindCharm ==1
       )
      ) {
    // 1) Monte Carlo
    if (ArteTable<MVER>::booked()) {
      ismk(28);
      ismksc(8.0);
      for( ArteTable<MVER>::iterator mverIt = ArteTable<MVER>::begin();
	   mverIt != ArteTable<MVER>::end();++mverIt){
	if( (*mverIt)->flag != 2 || (*mverIt)->z>50.0) continue;
	// Now find all mtras from this mver
	int ntr = 0;
	for (ArteRelationIterator<MTRA> mtraIt = (*mverIt)->etra.begin();
	     mtraIt != (*mverIt)->etra.end(); ++mtraIt) {
	  ArtePointer<RSEG> segment = RSEGCausedBy(*mtraIt);
	  if (segment != 0) if (segment->nhit >= 5) {
	    bool not_used = true;
	    // HACK =======================================
	    //check the track wasn't used in a primary
	    for (GroverParticleList::iterator ipar = allParticles->begin();
		 ipar != allParticles->end(); ++ipar) {
	      if (ipar->ParentRSEG() != segment) continue;
	      if (ipar->UsedInPrimary() != 0) not_used = false; 
	    }
	    // ============================================
	    if (not_used) ntr++;
	  }
	}
	// If less than two tracks, forget it.
	if (ntr>1) {
	  float local[5], global[5];
	  for (ArteRelationIterator<MTRA> mtraIt = (*mverIt)->etra.begin();
	       mtraIt != (*mverIt)->etra.end(); ++mtraIt) {
	    global[0]=(*mverIt)->x; 
	    global[1]=(*mverIt)->y; 
	    global[2]=(*mverIt)->z; 
	    acGlobalToLocal(local,global,1);
	    ArtePointer<RSEG> segment = RSEGCausedBy(*mtraIt);
	    if (segment != 0) {
	      GroverParticle temp(segment);
	      temp.GlobalToLocal(); //max
	      temp.PropagateBeginTo(local[2]);//max
	      if (true) draw_track(&temp,Cyan);
	    }
	  }
	  // then draw vertex if findable
	  xpm[0]=local[0];
	  ypm[0]=local[1];
	  ispmci(Cyan);
	  ismk(28);
	  ismksc(8.0);
	  ipm(1,xpm,ypm);
	}
      }
    }
    // 2) Reconstructed
    for (GroverVertexList::iterator iver = allSecondaries->begin();
	 iver != allSecondaries->end(); ++iver) {
      for (GroverParticleList::iterator ipar = iver->particles.begin();
	   ipar != iver->particles.end(); ++ipar) {
	ipar->PropagateBeginTo(iver->Z());
	// Now draw tracks at original positions!!!
	//draw_track(ipar,Magenta);
	if (ipar->Type()==2) {
	  ArtePointer<RSEG> the_rseg = ipar->ParentRSEG();
	  GroverParticle temp_par(the_rseg);
	  temp_par.GlobalToLocal(); //max
	  temp_par.PropagateBeginTo(iver->Z());
	  draw_track(&temp_par,Blue);
	}
      }
      draw_vertex(iver,Blue);
    }
  }

  return 0;

}
