#ifndef _GROVER_HH_
#define _GROVER_HH_    


/////////////////////////////////////////////////////////////////////////////
//
//
//         GROVER  - Generic Reconstruction Of VERtices
//	   ============================================
//
//
//                          -VERSION-                                   
static const char
GroverVersion[] =            "1.00";
//               
//
//	   Adapted from ROVER by T.Perschke/J.van Tilburg
//
//	   Authors: D.Emeliyanov, V.Kekelidze, I.Kisel, 
//                  S.Maschiocchi, M.Sang, S.Schaller
//
//	   Provides an flexible and extensible set of routines and 
//	   classes for HERA-B vertexing. Uniform interface between VT
//	   (T.Lohse), the still-developing Rover, and Arte. See the web 
//	   documentation at MPI Munich for more information:
//
//   	   http://afmp03.mppmu.mpg.de:8080/
//
//	   and follow the Data Analysis link.
//
//
//         Current responsible: Max Sang (max.sang@cern.ch)
//
//         Other people who may be able to help with queries:
//	   Ivan Kisel (ikisel@mppmu.mpg.de), 
//         Dmitry Emeliyanov (emelian@mail.desy.de)
//
// Modification Log:
//
//(YY-MM-DD) Person Modification(s) 
//
// 00-05-14  Max    First Version 0.01 - Attempt to get a common interface
//                  to VT and Rover. Some minor hacking of VT and major
//	            hacking of rover_primaries. No second_rove or 
// 		    graphics yet. Classes are reasonably complete but still
//      	    lots of work left to be done.
// 00-05-16  Max    Graphics added, removed RPGeoFit from GroverVertex and 
//                  put it back into GroverPrimaries. General feeling that I
//                  don`t want our classes to get fat. Could even strip
//		    out the flags and chi-squareds. These too are in
//		    danger of multiplying as more routines are written.
//		    If people need to store stuff they should use their own
//		    structures, not the Grover classes. second_rove added.
// 00-05-24  Max    Lots of changes and bug fixes. Graphics separated into
//                  its own module. Graphics added to GroverVTPrimaries().
//		    The 'particles' list of each returned vertex contains 
//		    particles with updated parameters - follow pointers for 
//		    original track parameters.
//		    Internal alignment in GroverPrimaries() works now :-)
// 00-05-26  Max    Added GroverPrimEval() - basically does the same sort of 
//                  thing as Verona by J.van Tilburg - calculates efficiencies
//		    etc for a pair of vertex lists (one real, one MC)
// 00-07-31  Max    v0.10 - Special version with all VT functionality 
//                  removed
// 00-09-07  Max    Bug removed (thanks Dmitry) in GroverVertex::SetSlope*()
// 00-09-15  Max    First version in Arte (1.00). VT functionality still 
//                  disabled!
// 00-09-26  Max    Some internal details in GroverVertex changed
// 00-09-27  Max    Added GroverParticle::GroverParticle(GroverVertex*)
// 00-10-01  Max    1) Added GroverAddBest(...) which adds best track in
//                     a list to the user's vertex.
//                  2) Changed default cuts to stop default calling of 
//                     secondary finders. GroverEvent(0) should now 
//                     have same effect as Rover primary finder.
/////////////////////////////////////////////////////////////////////////////



#ifndef _ARTE_H_
#define _ARTE_H_
#include "arte/arte.hh"
#endif

#ifndef _VECTOR_H_
#define _VECTOR_H_
#include <vector>
#endif

#ifndef _TFILE_H_
#define _TFILE_H_
#include "TFile.h"
#endif



class GroverParticle; // pre-definitions
class GroverVertex;
class GroverWire;



////////////////////////////////////////////////////////////////////////
// GroverParticleList
// ------------------
// std::vector of GroverParticles is good enough for the moment.
// It seems to provide pretty much all that we need.
// Indexing should work just like an array, as long as you don't
// do something silly like remove a particle from the middle of the
// list (this renumbers the later elements)
typedef vector<GroverParticle> GroverParticleList;



////////////////////////////////////////////////////////////////////////
// GroverVertexList
// ----------------
// std::vector again. Could be changed later if necessary
typedef vector<GroverVertex> GroverVertexList;



////////////////////////////////////////////////////////////////////////
// GroverWireList
// --------------
// As above
typedef vector<GroverWire> GroverWireList;



////////////////////////////////////////////////////////////////////////
// GroverParticle
// --------------
//
class GroverParticle 
{
public:
  // implicit copy constructor
  GroverParticle  (void);                    // Default constructor
  GroverParticle  (const ArtePointer<RSEG>); // Constructor for RSEGs
  GroverParticle  (const ArtePointer<RTRA>); // Constructor for RTRAs
  GroverParticle  (const ArtePointer<MTRA>); // Constructor for MTRAs
  GroverParticle  (GroverVertex*);           // Constructor for a vertex
 ~GroverParticle  (void);                    // Default destructor
  // Formatted dump of all private data in particle to stdout
  void  Dump(void);
  // Get coordinate system
  inline bool InLocal()  {return inlocal;}
  inline bool InGlobal() {return inglobal;}
  // Transform coordinate system
  int   LocalToGlobal(void);
  int   GlobalToLocal(void);
  // Simple smearing for MC particle (no effect if not MC)
  int   Smear(int);//NotDone!!//
  // already associated to a primary vertex?
  inline int UsedInPrimary(void) {return usedinprimary;}
  inline int SetUsedInPrimary(int m) {usedinprimary = m; return 0;}
  // already associated to a secondary vertex?
  inline int UsedInSecondary(void) {return usedinsecondary;}
  inline int SetUsedInSecondary(int m) {usedinsecondary = m; return 0;}
  // Weights  - use as you please if you write a new routine
  // *** I think these could be removed - Max
  inline float  PrimaryWeight(void) {return wprim;}
  inline int SetPrimaryWeight(float wprim_new) {wprim = wprim_new; return 0;}
  inline float  SecondaryWeight(void) {return wsec;}
  inline int SetSecondaryWeight(float wsec_new) {wsec = wsec_new; return 0;}
  // lowest Chi2 distance to a wire, particle, vertex or space point
  float Chi2DistTo(GroverWire*);
  float Chi2DistTo(GroverParticle*);
  float Chi2DistTo(GroverVertex*);
  float Chi2DistTo(float x, float y, float z);
  // shortest spatial distance to a wire, particle, vertex or space point (cm)
  float DistanceTo(GroverWire*);
  float DistanceTo(GroverParticle*);
  float DistanceTo(GroverVertex*);
  float DistanceTo(float x, float y, float z);
  // Propagate one end of the particle segment to a plane in z. 
  // The position and slope referred to by BeginX(), BeginSlopeX() etc. 
  // get replaced by the new values. The covariance matrix will, in 
  // general, also change. The propagation is not always meaningful!!!
  // Attempts to reverse the sense of 'beginning' and 'end' will fail
  int PropagateBeginTo(float z = 0.0);    
  int PropagateEndTo(float z = 2000.);//NotDone!!//
  // Get the original z positions of the start and end of track.
  // By projecting the particle to these planes one obtains access to all
  // of the original parameters.
  inline float OriginalBeginZ(void) {return ozb;}
  inline float OriginalEndZ(void) {return oze;}
  // Scalar momentum at start of track (0. if unknown)
  inline float Momentum(void) {return p;}
  // 1/Momentum (0. if unknown)
  inline float InvMomentum(void) {return invp;}
  // Mass of track (-999. if unknown)
  inline float Mass(void) {return mass;}       
  // x position of start of 'track'
  inline float BeginX(void) {return xb;}
  // y position of start of 'track'
  inline float BeginY(void) {return yb;}
  // z position of start of 'track'
  inline float BeginZ(void) {return zb;}
  // x position of end of 'track'
  inline float EndX(void) {return xe;}
  // y position of end of 'track'
  inline float EndY(void) {return ye;}
  // z position of end of 'track'
  inline float EndZ(void) {return ze;}
  // x slope @ start of 'track'
  inline float BeginSlopeX(void) {return axb;}
  // y slope @ start of 'track'
  inline float BeginSlopeY(void) {return ayb;}
  // x slope @ end of 'track'
  inline float EndSlopeX(void) {return axe;}
  // y slope @ end of 'track'
  inline float EndSlopeY(void) {return aye;}
  // i'th el. of cov. matrix at start of 'track'
  inline float BeginCov(int i) 
    {if (i>=0&&i<=20) return covb[i]; else return -999.;}
  // i'th el. of cov. matrix at end of 'track'
  inline float EndCov(int i)   
    {if (i>=0&&i<=20) return cove[i]; else return -999.;}
  // Number of hits in VDS
  inline int Nhit(void) {return nhit;}
  inline int SetNhit(int newnhit) {nhit=newnhit; return 0;}
  // Electric charge (-999 if unknown)
  inline int Charge(void) {return q;}
  inline int SetCharge(int newq) {q=newq; return 0;}
  // Index - label a particle with an integer 
  inline int Index(void) {return index;}
  inline int SetIndex(int i) {index=abs(i); return (i>=0) ? 0 : -1;}
  // Dist - label a particle with a float
  inline float Dist(void) {return dist;}
  inline int SetDist(float d) {dist=fabs(d); return 0;}
  // Constraints - if n != -1, the particle is participating in
  // mass constraint n (only meaningful if the particle is in
  // the internal particle list of a vertex!)
  inline int SetConstraint(int i) {constraint=i; return 0;}
  inline int GetConstraint(void) {return constraint;}
  // Chi2 contribution to a fit
  inline int SetChi2(float c) {chi2 = (c>=0.0) ? c:0.0; return 0;}
  inline float Chi2(void) {return chi2;}
  // 'Type of particle' flag tells you what was used to create
  //  this GroverParticle.
  //  This allows you to follow pointers
  //  back to the raw data - see manual.
  //  if(Type() == ...) the particle is a ...
  //  1 = RTRA entry (full track)
  //  2 = RSEG entry (subdetector segment only)
  //  3 = RCCL entry (photon / neutral hadron candidate)
  //  4 = MTRA entry (MC particle)
  // (5 - 49 reserved for future use)
  // 50 = another GroverParticle
  // 51 = A GroverParticleList
  // 52 = A GroverVertex
  inline int Type(void) {return type;}
  // Now some getters for pointers to parent
  // objects. Only one of them makes
  // sense: check Type() and then read the
  // one you need. All the others will
  // contain a null pointer anyway.
  inline ArtePointer<RTRA>   ParentRTRA(void) {return parRTRA;}
  inline ArtePointer<RSEG>   ParentRSEG(void) {return parRSEG;}
  inline ArtePointer<RCCL>   ParentRCCL(void) {return parRCCL;}
  inline ArtePointer<MTRA>   ParentMTRA(void) {return parMTRA;}
  inline GroverParticle*     ParentGP(void)   {return parGP;}
  inline GroverParticleList* ParentGPL(void)  {return parGPL;}
  inline GroverVertex*       ParentGV(void)   {return parGV;}
  // Now some 'direct access' setters. These should ONLY be used to *update*
  // particle parameters. USE WITH CARE!!!
  // Both SetMomentum and SetInvMomentum set p AND invp.
  // * THESE ARE TRICKY! Current algorithm projects the _original_ parameters
  // * so the updated parameters will be lost after a projection. Perhaps
  // * there is a case for a clever function here!!!! - Max
  inline int SetMomentum(float p_new)
    {if(p_new>0) {p=p_new; invp=1./p_new; return 0;} else return -1;}
  inline int SetInvMomentum(float invp_new)
    {if(invp_new>0) {invp=invp_new; p=1./invp_new; return 0;} else return -1;}
  inline int SetMass(float mass_new)
    {if(mass_new>=0) {mass=mass_new; return 0;} else return -1;}
  inline int SetBeginX(float xb_new)
    {xb = xb_new; return 0;}
  inline int SetBeginY(float yb_new)
    {yb = yb_new; return 0;}
  inline int SetBeginZ(float zb_new)
    {zb = zb_new; return 0;}
  inline int SetEndX(float xe_new)
    {xe = xe_new; return 0;}
  inline int SetEndY(float ye_new)
    {ye = ye_new; return 0;}
  inline int SetEndZ(float ze_new)
    {ze = ze_new; return 0;}
  inline int SetBeginSlopeX(float axb_new)
    {axb = axb_new; return 0;}
  inline int SetBeginSlopeY(float ayb_new)
    {ayb = ayb_new; return 0;}
  inline int SetEndSlopeX(float axe_new)
    {axe = axe_new; return 0;}
  inline int SetEndSlopeY(float aye_new)
    {aye = aye_new; return 0;}
  inline int SetBeginCov(float new_covb[21])
    {for(int i=0;i<21;i++) covb[i]=new_covb[i]; return 0;}
  inline int SetBeginCov(int index, float covb_new) // <- note syntax
    {if(index>=0&&index<=20) {covb[index]=covb_new; return 0;} else return -1;}
  inline int SetEndCov(float new_cove[21])
    {for(int i=0;i<21;i++) cove[i]=new_cove[i]; return 0;}
  inline int SetEndCov(int index, float cove_new)   // <- note syntax
    {if(index>=0&&index<=20) {cove[index]=cove_new; return 0;} else return -1;}
  // Reset the track parameters to their original values (e.g. after a fit)
  int ResetParameters(void);
  // Setters for updating pointers to parents. Use with EXTREME CARE!!!
  int SetParent(ArtePointer<RTRA>);
  int SetParent(ArtePointer<RSEG>);
  int SetParent(ArtePointer<RCCL>);
  int SetParent(ArtePointer<MTRA>);
  int SetParent(GroverParticle*);
  int SetParent(GroverParticleList*);
  int SetParent(GroverVertex*);
  // Setters for original parameters - use only when building a new particle
  // from scratch...
  inline int SetOBeginX(float oxb_new)
    {oxb = oxb_new; return 0;}
  inline int SetOBeginY(float oyb_new)
    {oyb = oyb_new; return 0;}
  inline int SetOBeginZ(float ozb_new)
    {ozb = ozb_new; return 0;}
  inline int SetOEndX(float oxe_new)
    {oxe = oxe_new; return 0;}
  inline int SetOEndY(float oye_new)
    {oye = oye_new; return 0;}
  inline int SetOEndZ(float oze_new)
    {oze = oze_new; return 0;}
  inline int SetOBeginSlopeX(float oaxb_new)
    {oaxb = oaxb_new; return 0;}
  inline int SetOBeginSlopeY(float oayb_new)
    {oayb = oayb_new; return 0;}
  inline int SetOEndSlopeX(float oaxe_new)
    {oaxe = oaxe_new; return 0;}
  inline int SetOEndSlopeY(float oaye_new)
    {oaye = oaye_new; return 0;}
  inline int SetOBeginCov(float new_ocovb[21])
    {for(int i=0;i<21;i++) ocovb[i]=new_ocovb[i]; return 0;}
  inline int SetOBeginCov(int index, float ocovb_new) // <- note syntax
    {if(index>=0&&index<=20) 
      {ocovb[index]=ocovb_new; return 0;} else return -1;}
  inline int SetOEndCov(float new_ocove[21])
    {for(int i=0;i<21;i++) ocove[i]=new_ocove[i]; return 0;}
  inline int SetOEndCov(int index, float ocove_new)   // <- note syntax
  {if(index>=0&&index<=20) 
    { ocove[index]= ocove_new; return 0;} else return -1;}
  inline int SetOMomentum(float new_p) 
    {op = new_p; return 1;}
  inline int SetGlobal(bool newg) {inglobal=newg; return 0;}
  inline int SetLocal(bool newl) {inlocal=newl; return 0;}
private:
  float xb;  // Current first and last points
  float yb;
  float zb;
  float axb;
  float ayb;
  float covb[21]; // x,y,z,ax,ay,p
  float xe;
  float ye;
  float ze;
  float axe;
  float aye;
  float cove[21];
  float oxb; // Original first and last points (private)
  float oyb;
  float ozb;
  float oaxb;
  float oayb;
  float ocovb[21];
  float oxe;
  float oye;
  float oze;
  float oaxe;
  float oaye;
  float ocove[21];
  float p;
  float op;
  float invp;
  float dinvp;
  float mass;
  int   nhit;
  int   q;
  int   type;
  int   usedinprimary;
  int   usedinsecondary;
  int   index;
  ArtePointer<RTRA>   parRTRA; // - could probably be clever and use one
  ArtePointer<RSEG>   parRSEG; // pointer with casts, but I'm not risking
  ArtePointer<RCCL>   parRCCL; // it :-) Stick with a 'parent type' flag
  ArtePointer<MTRA>   parMTRA; // and a handful of redundant pointers.
  GroverParticle*     parGP;
  GroverParticleList* parGPL;
  GroverVertex*       parGV;
  bool  inlocal;
  bool  inglobal;
  float wprim;
  float wsec;
  float dist;
  float chi2;  // chi2 contribution to latest un-mass-constrained fit
  float chi2_m;// chi2 contribution to latest mass-constrained fit
  int   constraint; // took part in constraint number N
  float z1_closest;
  float z2_closest;
};



////////////////////////////////////////////////////////////////////////
//  GroverVertex
//  ------------
//
class GroverVertex {
public:
  GroverVertex();
  GroverVertex(GroverParticle*);                                 // 1
  GroverVertex(GroverParticle*,GroverParticle*);                 // 2
  GroverVertex(GroverParticle*,GroverParticle*,GroverParticle*); // 3
  GroverVertex(GroverParticle*,GroverParticle*,GroverParticle*,  // 4
	       GroverParticle*);                                 
  GroverVertex(GroverParticle*,GroverParticle*,GroverParticle*,  // 5
	       GroverParticle*,GroverParticle*);
  GroverVertex(GroverParticle*,GroverParticle*,GroverParticle*,  // 6
	       GroverParticle*,GroverParticle*,GroverParticle*);
  GroverVertex(GroverParticleList*);                             // N
  GroverVertex(ArtePointer<RVER>);
  GroverVertex(ArtePointer<MVER>);
  // Construct from a point in LOCAL coordinates
  GroverVertex(const float xx, const float yy, const float zz);
  // Construct from a VT vertex index (with or without corresponding GPL)
  GroverVertex(int index, GroverParticleList* gpl);
  GroverVertex(int index);
 ~GroverVertex();
  // Dump private members to stdout
  void Dump(void);
  // Index - label a vertex with an integer
  inline int Index(void) {return index;}
  inline int SetIndex(int i) {index=abs(i); return (i>=0)?0:-1;}
  // Number of tracks in internal particle list
  inline int Size(void) 
    {if (!particles.empty()) return particles.size(); else return 0;};
  // Sort particle list by their spatial distance in x-y plane to vtx (x,y)
  int SpatialSort(void);
  // count the good quality tracks. Not done - should I?
  int   RefSize(void); 
  // Add a particle to internal particle list
  int   Add(GroverParticle*);     
  // Add everything in another GPL (if empty: returns -1 and does nothing)
  int   Add(GroverParticleList*);
  // Remove particle number i (starts at 0) or remove by pointer
  // (standard STL method) - this clears *all* constraint information
  int   Remove(int i);
  int   Remove(GroverParticle* ptr);
  // Add a mass constraint, with a list of two or more particles (max 10)
  // to which this new constraint applies. Mass constraint can be given
  // as a float or a TOPP pointer. The particles can be given as integer
  // indices in the particles list or as pointers to particles which are
  // already in the list (pointers to particles outside the list will
  // return an error)
  int   AddConstraint(float mass, int p1, int p2, 
		      int p3=-1, int p4=-1, int p5=-1, int p6=-1, 
		      int p7=-1, int p8=-1, int p9=-1, int p10=-1);
  int   AddConstraint(float mass, GroverParticle* p1, GroverParticle* p2, 
		      GroverParticle* p3=0, GroverParticle* p4=0, 
		      GroverParticle* p5=0, GroverParticle* p6=0, 
		      GroverParticle* p7=0, GroverParticle* p8=0, 
		      GroverParticle* p9=0, GroverParticle* p10=0);
  int   AddConstraint(ArtePointer<TOPP>, int p1, int p2, 
		      int p3=-1, int p4=-1, int p5=-1, int p6=-1, 
		      int p7=-1, int p8=-1, int p9=-1, int p10=-1);
  int   AddConstraint(ArtePointer<TOPP>, GroverParticle* p1, GroverParticle* p2, 
		      GroverParticle* p3=0, GroverParticle* p4=0, 
		      GroverParticle* p5=0, GroverParticle* p6=0, 
		      GroverParticle* p7=0, GroverParticle* p8=0, 
		      GroverParticle* p9=0, GroverParticle* p10=0);
  // Remove all mass constraints
  int   ClearConstraints(void);
  // Perform a vertex fit with mass constraint m (m <=0: unconstrained fit). Not
  // suitable for primary vertices.
  int   Fit(float m);
  // Vertex fit with mass constrained to a particle in the TOPP table (eg J/psi = 295)
  int   Fit(ArtePointer<TOPP> hypothesis);
  // Fit which is constrained or unconstrained according to current
  // constraint flags
  int   Fit(void);
  // chi-squared of the last fit (0. if undefined)
  inline float Chi2(void) {return chi2;} 
  inline int SetChi2(float newc) {chi2=newc; return 0;}
  // mass of the vertex (0. if unknown)
  inline float Mass(void) {return m;} 
  // error on the vertex mass (0. if unknown)
  inline float MassError(void) {return dm;} 
  // give estimate of the vertex mass as a starting point for a fit
  inline int SetMass(float newm) {m=fabs(newm); return 0;}
  inline int SetMassError(float newdm) {dm=fabs(newdm); return 0;}
  // Position of vertex
  inline float X() {return x;}
  inline float Y() {return y;}
  inline float Z() {return z;}
  // Manually set the position of the vertex in the current coordinate system
  inline int SetX(float xx) {x=xx; return 0;}
  inline int SetY(float yy) {y=yy; return 0;}
  inline int SetZ(float zz) {z=zz; return 0;}
  // Slope and momentum of vertex
  inline float Momentum(void) {return p;}
  inline int SetMomentum(float newp) {p=fabs(newp);return 0;}
  inline float SlopeX(void) {return ax;}
  inline float SlopeY(void) {return ay;}
  inline int SetSlopeX(float newax) {ax=newax;return 0;}
  inline int SetSlopeY(float neway) {ay=neway;return 0;}
  // Manually set/get the covariance matrix
  inline float Cov(int i) {if(i>=0&&i<36) return cov[i]; else return -999.;}
  inline int SetCov(float value[36])
    {for (int i=0;i<36;i++) cov[i] = value[i]; return 0;}
  inline int   SetCov(int element, float value)
    {if(element>=0&&element<36){cov[element]=value;return 0;} else return -1;}
  // Estimate a diagonal CM using the particles in the internal list.
  // The diagonal elements can be scaled for increased freedom.
  // The z covariance may also be set (default 0.01 means sig_z = 1mm)
  int   EstimateCov(float scale=1., float zcov=0.01); 
  // Coordinate transformations between VDS system and HERA-B system.
  // - use with care! they might be wrong, especially covariances :-|
  // The particles in the internal list are also transformed and the 
  // same caveat applies to their covariances.
  int   LocalToGlobal(void);
  int   GlobalToLocal(void);
  // Get coordinate system
  inline bool InLocal(void)  {return inlocal;}
  inline bool InGlobal(void) {return inglobal;}
  // Estimate scalar momentum using particles in list (a la VT)
  float EstimateMomentum(void);
  // Set and get wire number with which the vertex is associated
  inline int Wire(void) {return wnr;}
  inline int SetWire(int wire) 
    {if (wire>0&&wire<9) {wnr=wire; return 0;} else return -1;}
  // Spatial and chi2 distances to a wire, vertex, particle, or space point
  float DistanceTo(GroverParticle*);
  float Chi2DistTo(GroverParticle*);
  float DistanceTo(GroverVertex*);
  float Chi2DistTo(GroverVertex*);
  float DistanceTo(GroverWire*);
  float Chi2DistTo(GroverWire*);
  float DistanceTo(float xx, float yy, float zz);
  float Chi2DistTo(float xx, float yy, float zz);
  // *** PUBLIC VARIABLES *** (beg for forgiveness!)
  // particles and trashcan
  // are public but don't add or remove particles please! Adding is done using 
  // Add() and removing is not implemented yet. If you have suggestions about 
  // Remove(...) please email the authors. The trashcan is available
  // immediately after an operation such as fitting, but by default is cleared
  // by other member functions before they start.
  GroverParticleList particles;
  GroverParticleList trashcan;
  // Calculate x,y distances to vertex 
  int UpdateDistances(void);
  // Reset parameters of every track in the particle and trashcan lists
  int ResetParticles(void);
  // Vika's stuff: under test
  inline int SetBx(float bxbx){bx=bxbx;return 0;}
  float Bx(void){return bx;}
  inline int SetBy(float byby){by=byby;return 0;}
  float By(void){return by;}
  inline float DecayLength(void){return decay;}
  inline int SetDecayLength(float len){decay=len; return 0;}
  inline float GCov(unsigned int n){if (n>=0&&n<gcov.size()) return gcov[n]; else return 1e9;}
  // testing:
  void  dump_cons(void);
  /*
  inline float GCov(int i, int j)
    {if (i<0||j<0||i>=GCOVDIM||j>=GCOVDIM) return -1e10; 
     else return gcov[i][j];}
  inline int SetGCov(int i, int j, float gcov_ij)
    {if (i<0||j<0||i>=GCOVDIM||j>=GCOVDIM) return -1;
     else {gcov[i][j]=gcov_ij;
           return 0;
     }}
  */
private:
  static const int COVDIM  = 36; // x,y,z,ax,ay,p,bx.by
  static const int CONSTDIM= 10; // maximum number of constraints
  static const int MAXTRKS = 15; // max. # tracks for *constrained* fits
  int   index;
  float x;
  float y;
  float z;
  float ax;
  float ay;
  float p;
  float m;
  float dm;
  bool  inlocal;
  bool  inglobal;
  float cov[COVDIM];
  float wprim;
  float wsec;
  int   wnr;
  int   refntrincluster;
  float chi2;
  bool  fitted;
  int   n_constraints;
  float constr_mass[CONSTDIM];
  int   constr_map[CONSTDIM][MAXTRKS];
  // Vika's:
  float bx; // x component of impact parameter at z=0
  float by;
  float decay;
  vector<float> gcov;
};




///////////////////////////////////////////////////////////////////////
//  GroverWire
//  ----------
//  
class GroverWire {
public:
  // default constructor -> vertical wire at 0,0,0 in global coordinates
  GroverWire();                 
  // construct from a GTAR entry (Assumes GTAR in global coordinates!)
  GroverWire(ArtePointer<GTAR>);
  // default destructor
 ~GroverWire();                  
  // formatted dump of values to stdout
  void Dump(void);
  // Wire number
  inline int Number(void) {return wnr;}
  // x position of wire centre
  inline float X(void) {return x;}          
  inline int SetX(float xx) {x=xx; return 0;}
  // y position of wire centre
  inline float Y(void) {return y;}           
  inline int SetY(float yy) {y=yy; return 0;}    
  // z position of wire centre
  inline float Z(void) {return z;}           
  inline int SetZ(float zz) {z=zz; return 0;}  
  // dimension of wire in x 
  inline float SizeX(void) {return dx;}         
  inline int SetSizeX(float ddx) {if (ddx>0){dx=ddx;return 0;} else return -1;}
  // dimension of wire in y 
  inline float SizeY(void) {return dy;}     
  inline int SetSizeY(float ddy) {if (ddy>0){dy=ddy;return 0;} else return -1;}
  // dimension of wire in z
  inline float SizeZ(void) {return dz;}       
  inline int SetSizeZ(float ddz) {if (ddz>0){dz=ddz;return 0;} else return -1;}
  // wire shape (cylinder, band etc)
  inline int   Shape(void) {return shape;}
  // 1 if horizontal wire, 0 otherwise
  inline int   Horizontal(void) {return horizontal;}
  // 1 if vertical wire, 0 otherwise
  inline int   Vertical(void) {return vertical;}
  // Rotation matrix, wire system to lab system
  inline float Rot(int i) {if(i>=0&&i<9) return rot[i]; else return -999.;}
  // Number of particles associated to this wire (don't use blindly :-)
  inline int   NParticles(void) {return ntr;}
  inline int   IncrementNParticles(void) {return ++ntr;}
  inline int   DecrementNParticles(void) {return --ntr;}
  inline int   SetNParticles(int nn) {ntr = abs(nn); return 0;}
  // Number of primary vertices associated to this wire (-"-)
  inline int   NPrimaries(void) {return nprim;}
  inline int   IncrementNPrimaries(void) {return ++nprim;}
  inline int   DecrementNPrimaries(void) {return --nprim;}
  inline int   SetNPrimaries(int nn) {nprim = abs(nn); return 0;}
  // Number of secondary vertices associated to this wire (-"-)
  inline int   NSecondaries(void) {return nsec;}
  inline int   IncrementNSecondaries(void) {return ++nsec;}
  inline int   DecrementNSecondaries(void) {return --nsec;}
  inline int   SetNSecondaries(int nn) {nsec = abs(nn); return 0;}
  // Coordinate transformations between VDS system and HERA-B system
  int GlobalToLocal(void);
  int LocalToGlobal(void);
private:
  int   wnr;
  float x;
  float y;
  float z;
  float dx;
  float dy;
  float dz;
  int   shape;
  int   horizontal;
  int   vertical;
  float rot[9];
  bool  intarget;// target wire coordinates
  bool  inlocal; // vds local coordinates
  bool  inglobal;// global HERA-B coordinates
  int   ntr;
  int   nprim;
  int   nsec;
};




///////////////////////////////////////////////////////////////////////
//  GroverCuts
//  ----------
//
class GroverCuts {
public:
  GroverCuts();    // default constructor - sets defaults!
 ~GroverCuts();    // default destructor
  void Dump(void); // formatted dump of current values to stdout
  //
  //     0) General cuts
  //
  // Vertexing package for primary finding - 1=Grover, 2=VT (default 1)
  int   VertexingPackage;
  // Event display with HIGZ (default 0 = off)
  int   GroverDraw;
  // Which type of secondaries to search for? (defaults 1,1,1)
  int   FindJpsiPlusX;
  int   FindUnbiased;
  int   FindCharm;
  // Print out messages all over the place
  int   Verbose;
  // Use the evaluation package (Efficiency calculations etc)
  int   GPEActivate;
  // use HBOOK/PAW and/or ROOT in GPE (default Paw=0, Root=1)
  int   Paw;
  int   Root;
  //
  //
  //     1) Cuts for MC studies
  //
  //
  // Mass constraint to use in fit
  float MassConstraint;
  // insert real m and/or p into tracks for vertex resolution studies (def. 0,0)
  int   CheatMass;
  int   CheatMomentum;
  // smear momenta by these factors (set to zero for no smearing)
  // const = constant term (all momenta), var = p dependent term
  // (defaults 0.0,0.0; TDR values = 0.0045,0.00005)
  int   CheatSmearing;
  float CheatSmearConst;
  float CheatSmearVar;
  //
  //
  //     2) Cuts for VT - described in T.Lohse's manual
  //        Used to be parameters, now they are cuts 
  //        - VT modified by M.Sang 11/05/00
  // 
  // PARTICLE MASSES - please substitute the TOPP values if possible!
  double VTmPsi;
  double VTmKs;
  double VTmB;
  double VTmpi;
  double VTmmu;  
  double VTme;
  // max chi2 distance for trk/wire assignment. (default 10.)
  float VTch2wtr;
  // max chi2 distance for B0/wire assignment. (default 20.)
  float VTch2wB;
  // parameters for clustering tracks which have been assigned to a common
  // wire. VTxlmin is the minimum hit separation along a wire and VTxlwid
  // is the maximum width of a cluster. The clusters are used as seeds for 
  // primary vertices (default VTxlmin=.015, VTxlwid=.050)
  float VTxlmin;
  float VTxlwid;
  // min trks on a primary - ignored if it's a B0 primary, which can even
  // have 0 tracks left over (i.e. just B0 intersection with a wire)
  // (default 3)
  int   VTntmin;
  // max chi2 distance for a track to a primary vtx seed for trk/vtx merging 
  // (default 14.)
  float VTch2vtr;
  // max chi2 for merging 2 trks into a new secondary (default 6.)
  float VTch2stt;
  // max chi2 for adding a track to a secondary vertex (default 8.)
  float VTch2stv;
  // min chi2 distance of a secondary vertex from a wire (default 1.)
  float VTch2svw;
  // max impact parameter in (cm) of track to the B0 secondary if the 
  // search is restricted to the vicinity of the B0 (eg for charm decays)
  // (default 0.1)
  float VTBvimpc;
  // max # of primary vtx tracks assignable to a secondary (default 1)
  int   VTnmix;
  // max chi2 distance of 2 vertices for vertex merging (default 6.)
  float VTch2vv;
  // Set to 999 before first call to VT to override default cuts (default 999)
  int   VTtweak;
  // Event display with HIGZ (default 0 = off)
  int   VTDraw;
  //
  //
  //      3) Cuts for the Rover Primary vertex finder
  //
  //
  // Which tracks to use? 1=Rover (default) 2=all VDS RSEGs 3=all RTRAs
  int   RPTrackChoice;
  // How many VDS hits (minimum) in the tracks? (default 5)
  int   RPTrackQualNH;
  // minimum number of tracks in a primary (default 3)
  int   RPMinTrksInVtx;
  // maximum number of sigma for associating tracks to wires 
  // or each other (default 3.5)
  float RPTrkWireSigmaMax;
  // 
  //
  //      4) Cuts for the GroverPrimEval vertex finder evaluation
  //         routine (based on VERONA by J.van Tilburg)
  //
  //
  // Only print the stuff every Nth event (default = 100)
  int   GPEUpdateFreq;
  // Minimum # hits for a track to be counted
  int   GPEMinHitsInTrack;
  // Minimum # tracks for a vertex to be counted
  int   GPEMinTracksInVertex;
  // Maximum chi2 for a vtx to be accepted
  float GPEMaxChi2;
  // Max chi2 per track for a vtx to be accepted
  float GPEMaxChi2PerTrack;
  // Size of windows (in cm) for matching real vertices with 
  // MC vertices (defaults 0.0200, 0.4000 respectively)
  float GPETransverseWindow;
  float GPELongitudinalWindow;
  // Chi (sqrt(chi2)) cut for same (default 5.0)
  float GPEChiWindow;
  //
  //
  //      5) Cuts for the Grover secondary vertex finders
  //
  //
  // How many VDS hits (minimum) in the tracks? (default 7)
  int   RSTrackQualNH;
  // J/psi seed trks: trk-trk chi2 dist (def. 20) and spatial dist (def. 0.1)
  float RSJpsiSeedMaxChi2;
  float RSJpsiSeedMaxDist;
  /*
  // Minimum chi2 separation of secondary candidate and another vertex
  // (secondary or primary) for the candidate to be accepted as detached
  // (deafult 3.0)
  float RSVtxVtxMinChi2;
  // Maximum Chi2 and spatial distances for a track to be added to a
  // secondary candidate (defaults 5.0 and 0.2, respectively) 
  float RSTrkVtxMaxChi2;
  float RSTrkVtxMaxDist;
  // Minimum (spatial distance of track to secondary
  //          / spatial distance of track to closest primary)
  // (default 1/3)
  float RSTrkVtxDistRatio;
  */
  // Minimum impact parameter significance for a seed track (default 10.0)
  float RSMinSeedIPS;
  // Max chi2 contribution of a track to a 
  // secondary (otherwise rejected) (def. 5.0)
  float RSMaxChi2Contrib;
  // Maximum chi2 distance of trk to vtx to consider refitting (def. 10)
  float RSMaxTrkVtxChi2;
  // Maximum chi2 dist of two tracks to consider fitting as a seed (def. 10)
  float RSMaxTrkTrkChi2;
  // Maximum chi2 of a two-track seed after initial fit (default 10)
  float RSMax2TVtxChi2;
  // Maximum chi2/dof to accept a secondary (def. 5.0)
  float RSMaxChi2PDOF;
private:
  // none! Yes, it could probably have been a struct.. ;-)
};




/////////////////////////////////////////////////////////////////////////
//  Other Stuff
//  -----------
//

// Colours for the HIGZ graphics
const int Black        =  1,
          DarkGray     =  2,
          Blue         =  3,
          Green        =  4,
          Cyan         =  5,
          Red          =  6,
          Magenta      =  7,
          Brown        =  8,
          LightBlue    =  9,
          LightGreen   = 10,
          LightCyan    = 11,
          LightRed     = 12,
          LightMagenta = 13,
          Yellow       = 14,
          LightGray    = 15,
          White        =  0;

// If this is (re)set to 1, the targets are aligned using track positions
// only (ignoring vertices) at the beginning of the event.
extern int rover_newgeometry;

// Wire geometry for alignment
extern float wire_xpos[9],
             wire_ypos[9],
             wire_zpos[9],
             wire_xsize[9],
             wire_ysize[9],
	     wire_zsize[9];

// Configuration
extern GroverCuts groverCuts;
extern TFile* GroverOutputFile;




//////////////////////////////////////////////////////////////////////////
//  Function Prototypes
//  -------------------
//

// External
extern int  acLocalToGlobal(float* local, float* global, int);
extern int  acGlobalToLocal(float* local, float* global, int);

// Configuration (usinit) and cleanup (usstop)
extern int  GroverConfig(char* filename);
extern void GroverStop(void);

// Filling to/from ARTE tables
extern int  GroverFillFromRSEG(GroverParticleList*);
extern int  GroverFillFromRTRA(GroverParticleList*);
extern int  GroverFillLikeRover(GroverParticleList*);
extern int  GroverFillFromGTAR(GroverWireList*);
extern int  GroverFillFromRVER(GroverVertexList*);
extern int  GroverFillFromMVER(GroverVertexList*,int type=1);// 1=primaries etc
extern int  GroverAddToRVER(GroverVertexList*, int type=1);  // -"-

// Graphics
extern int  graph_ini(void);
extern void draw_targets(GroverWireList*);
extern void draw_mc(float z_plane);
extern void draw_track(GroverParticle*,int colour);
extern void draw_vertex(GroverVertex*,int colour);

/*
// VT specific
extern int  GroverFillVTInputs(GroverParticleList*,GroverWireList*,
			       GroverCuts*);
extern int  GroverGetVTPrimaries(GroverVertexList*,GroverParticleList*,
				 GroverCuts*);
extern int  GroverVTPrimaries(GroverParticleList*,GroverWireList*,GroverCuts*,
			      GroverVertexList* primaries);
*/

// Grover specific
extern int  GeoFit(GroverVertex*);
extern void GroverSecondaryFit (GroverVertex*);
extern int  GroverQuickFit(float& x,  float& y,  float& z,
			   float& dx, float& dy, float& dz, 
			   float& chi2, float& mass,
			   ArtePointer<RTRA> rtra1, ArtePointer<RTRA> rtra2,
			   ArtePointer<RTRA> rtra3=0, ArtePointer<RTRA> rtra4=0, 
			   ArtePointer<RTRA> rtra5=0);
extern int  GroverQuickFit(float& x,  float& y,  float& z,
			   float& dx, float& dy, float& dz,
			   float& chi2, float& mass,
			   ArtePointer<RSEG> rseg1, ArtePointer<RSEG> rseg2, 
			   ArtePointer<RSEG> rseg3=0, ArtePointer<RSEG> rseg4=0, 
			   ArtePointer<RSEG> rseg5=0);
//extern int  GroverConstrainedFit(GroverVertex*, 
//				 vector< ArtePointer<TOPP> >*  masses);
extern int  GroverConstrainedFit(GroverVertex*, float mass);
extern void four_dim_inv(double any_matrix[4][4]); 
extern void five_dim_inv(double any_matrix[5][5]);
extern int  GroverCatsExtrap(ArtePointer<RSEG>,float,float*,float*,float*,
			     float*,float*,float*,float*);
extern int  GroverPrimaries(GroverParticleList*,GroverWireList*,GroverCuts*,
			    GroverVertexList* primaries);
extern int  GroverEvent(GroverCuts* optional_cuts);
extern int  GroverSecondaries(GroverParticleList* all,GroverWireList*,
			      GroverCuts*,GroverVertexList* input_primaries,
			      GroverVertexList* output_secondaries);
extern int  GroverUnbiasedSecondaries(GroverParticleList*,
				      GroverWireList*,GroverCuts*,
				      GroverVertexList* input_primaries,
				      GroverVertexList* output_secondaries);
extern int  GroverJpsiSec(GroverParticleList*,GroverWireList*,GroverCuts*,
			  GroverVertexList* primaries,GroverVertexList* jpsis);
extern int  GroverCharm(GroverParticleList*,GroverWireList*,GroverCuts*,
			GroverVertexList* primaries,GroverVertexList* charm);
extern int  GroverFindJpsis(GroverParticleList*,GroverCuts*,GroverVertexList*);
extern int  GroverSecFinder(GroverVertexList* mc,GroverVertexList* reco,
			    int type = 1, GroverCuts* c = 0);
extern int  GroverSecTest(void);
// Utilities
extern int  GroverPrimEval(GroverVertexList* reconstructed,
			   GroverVertexList* optional_mc,GroverCuts*);
extern int  PrimEvalInitPaw(void);
extern int  PrimEvalInitRoot(void);
extern bool SortGPByDist(GroverParticle,GroverParticle); // sorting functions
extern bool RevSortGPByDist(GroverParticle,GroverParticle);
extern bool SortGPByChi2(GroverParticle,GroverParticle);
extern bool RevSortGPByChi2(GroverParticle,GroverParticle);
extern bool SortGPBySecWeight(GroverParticle,GroverParticle);
extern bool RevSortGPBySecWeight(GroverParticle,GroverParticle);
extern bool SortGPByPrimWeight(GroverParticle,GroverParticle);
extern bool RevSortGPByPrimWeight(GroverParticle,GroverParticle);
extern void three_dim_inv(double a[3][3]);// quick matrix inversion
extern void four_dim_inv(double a[4][4]);
extern void five_dim_inv(double a[5][5]);
extern void six_dim_inv(double a[6][6]);
extern void seven_dim_inv(double a[7][7]);
extern ArtePointer<RSEG> RSEGCausedBy(ArtePointer<MTRA>);
extern ArtePointer<RTRA> RTRACausedBy(ArtePointer<MTRA>);
extern ArtePointer<MTRA> MTRAWhichCaused(ArtePointer<RSEG>);
extern ArtePointer<MTRA> MTRAWhichCaused(ArtePointer<RTRA>);
extern int  GroverRefitRSEG(GroverCuts* c = 0);

#endif

