// See high_pt.hh for details.

// Author: Vladislav Balagura (balagura@hera-b.desy.de, 16.05.2000)

#include "ptset/high_pt.hh"
#include "ptset/global_error.hh"
#include "ptset/setup_tools.hh"
#include "ptset/setup_menu.hh"

HPT_Setup hpt_setup;

/* The following is needed to make interactive command /create
   which reads the chamber names in the following format:
     /create{ {{i1- i1+}{o1- o1+}}
              {{i2- i2+}{o2- o2+}}
	      {{i3- i3+}{o3- o3+}} }
   This means: first superlayer {{i1- i1+}{o1- o1+}} is composed of 2 parts,
   part one (inner high-pt) has chambers i1-, i1+; part two (outer) - o1-, o1+;
   and so on for the 2d and 3d superlayers.
   */   
enum Result {NEXT, END, ERROR};
Result get_block(istream& is, string* s) { // reads block { ... }
  // taking into account the presence of nested blocks: { .. {..} .. }
  char c;
  if ( !(is >> c) ) return END;
  if ( c != '{' ) {
    global_error()("can not read \'{\' in front of chamber names"); 
    return ERROR;
  }
  int level = 1;
  *s = "";
  while ( 1 ) {
    is.get(c);
    if ( !is ) {
      global_error()("closing \'}\' is missing");
      return ERROR;
    }
    if ( c == '{' ) ++level;
    if ( c == '}' ) --level;
    if ( level == 0 ) break;
    (*s) += c;
  }
  return NEXT;
}
// this structure will be used to make an interactive command.
// It reads the input stream, fills vector<vector<vector<string> > > names
// (names[ilayer][ipart][ichamber])
// and then calls HPT_Setup::create_chambers (see below)
struct Get_HPT_Chamber_Names {
  HPT_Setup* setup;
  Get_HPT_Chamber_Names(HPT_Setup* fSetup) : setup(fSetup) {}
  void operator()(istream* is) {
    vector<vector<vector<string> > > names;
    string s1;
    while ( 1 ) {
      Result r = get_block(*is, &s1);
      if ( r == END ) break;
      else if ( r == ERROR ) {
	global_error() << "wrong format in the block of High-Pt chamber names";
	return;
      }      
      names.push_back(vector<vector<string> >());
      istrstream is1(s1.c_str());
      string s2;
      while ( 1 ) {
	Result r1 = get_block(is1, &s2);
	if ( r1 == END ) break;
	else if ( r1 == ERROR ) {
	  global_error()<< "wrong format in the block of High-Pt chamber names";
	  return;
	}      
	names.back().push_back(vector<string>());
	istrstream is2(s2.c_str());
	string name;
	while ( is2 >> name ) names.back().back().push_back(name);
      }
    }
    /*
    for (unsigned int il=0; il<names.size(); ++il) {
      cout << "il=" << il << '\n';
      for (unsigned int ip=0; ip<names[il].size(); ++ip) {
	cout << "ip=" << ip << '\n';
	for (unsigned int ich=0; ich<names[il][ip].size(); ++ich)
	  cout << "ich=" << ich << " " << names[il][ip][ich] << " ";
	cout << '\n';
      }
      cout << '\n';
    }
    */
    setup->create_chambers(names);
  }
};
// This function creates chambers distributed among layers and parts
// according to the given vector<vector<vector<string> > > names
void HPT_Setup::create_chambers(const vector<vector<vector<string> > >& names) {
  mIChambers.resize(names.size());
  for (unsigned int il=0; il<names.size(); ++il) {
    mIChambers[il].resize(names[il].size());
    for (unsigned int ip=0; ip<names[il].size(); ++ip) {
      mIChambers[il][ip].resize(names[il][ip].size());
      for (unsigned int ich=0; ich<names[il][ip].size(); ++ich) {
	pair<iterator, bool> p = insert(value_type(names[il][ip][ich],
						   HPT_Chamber()));
	if ( p.second == false ) {
	  global_error()("chamber name " + names[il][ip][ich] +
			 " is given twice");
	  return;
	}
	mIChambers[il][ip][ich] = p.first;
	mID[names[il][ip][ich]] = HPT_ChamberID(il, ip, ich);
      }
    }
  }
}
	  
// This command uses SetupMenu<HPT_Setup> with only one difference:
// the /create command in this menu is substituted by the command described
// above: it creates not just a collection of chambers without any special 
// ordering, but the chambers which are put into the superlayers and into
// "parts" : inner and outer parts.
void HPT_Setup::read(istream& is) {
  SetupMenu<HPT_Setup> menu(&hpt_setup);
  vector<vector<vector<string> > > names;
  menu("create") = make_interactive_command(Get_HPT_Chamber_Names(this));
  menu(&is);
  if ( global_error() ) return;
  
  //  cross_reference();
}


/* the following function 
     HPT_Setup::cross_reference()
   finds layers/parts automatically using z oordinate of the chambers. Thus it 
   takes just the collection of chambers (without layers/parts information)
   and structures them into layers/parts. It works
   correctly in the design case, when all chambers are installed, outer part is
   shifted from inner downstream in all layers. However it is not used in order
   to give more flexibility in assigning chambers to layers/parts. Currently
   this can be done "by hand" in hpt_geo.dat file using the /create command
   described above.
   */
// LessDouble_Fine is just LessDouble with epsilon = 1.0
// ("Fine" is in comparison with epsilon=20. which will be used later)
struct LessDouble_Fine {
  bool operator()(double x1, double x2) const { return x1 < x2 - 1.; }
};
void HPT_Setup::cross_reference() {
  double epsilon_coarse = 20.; // to distinguish different layers
  // double epsilon_fine = 1.;  
  // which used in LessDouble_Fine above is to distinguish inner/outer inside
  //  layer

  typedef map<double, vector<HPT_Setup::IChamber>, LessDouble_Fine> Zmap1;
  typedef map<double, Zmap1, LessDouble> Zmap;
  LessDouble ld_coarse(epsilon_coarse);
  Zmap zmap(ld_coarse);

  mIChambers.clear();
  mID.clear();
  for (HPT_Setup::IChamber ich = begin(); ich != end(); ++ich) {
    double z = (*ich).second.z;
    zmap[z][z].push_back(ich); // [coarse_group][fine_group]
  }
  
  mIChambers.resize(zmap.size());
  unsigned int il = 0;
  for ( Zmap::iterator iter=zmap.begin(); iter!=zmap.end(); ++iter, ++il ) {
    Zmap1& zmap1 = (*iter).second;
    mIChambers[il].resize(zmap1.size());
    unsigned int ipart = 0;
    for ( Zmap1::iterator iter1=zmap1.begin(); iter1!=zmap1.end();
	  ++iter1, ++ipart ) {
      vector<HPT_Setup::IChamber>& v = (*iter1).second;
      mIChambers[il][ipart] = v;
      for (unsigned int i_ch=0; i_ch<v.size(); ++i_ch) {
	HPT_Setup::IChamber ich = v[i_ch];
	mID[(*ich).first] = HPT_ChamberID(il, ipart, i_ch);
      }
    }
  }
}
  
void HPT_Setup::clear() {
  Setup<HPT_CellT, HPT_SectionT, Dummy, HPT_SetupT>::clear();
  mIChambers.clear();
  mID.clear();
}

/* Modification made on 12.09.00: wrting/reading the hits */
#include "ptset/fired_cell_tools.hh"
#include <zlib.h> // compression library

inline void write_short_int(short int i, string& s) {
  unsigned char b2 = i & 255;
  unsigned char b1 = i >> 8;
  s += b1; s += b2;
}
inline void write_short_int(short int i, ostream& os) {
  unsigned char b2 = i & 255;
  unsigned char b1 = i >> 8;
  os << b1 << b2;
}

inline unsigned short int read_short_int(istream& is) {
  unsigned char b2;
  unsigned char b1;
  is.get(b1).get(b2);
  return ((unsigned short int)b1<<8) | b2;  
}
inline unsigned short int read_short_int(unsigned char* buffer,
					 unsigned int* position) {
  unsigned short int result = (unsigned short int)buffer[*position]<<8;
  ++(*position);
  result |= buffer[*position];
  ++(*position);
  return  result;
}

// -------------------- I/O with compression --------------------
static unsigned char* small_buffer = 0;
static unsigned char* large_buffer = 0;
static unsigned long int small_buffer_length = 0;
static unsigned long int large_buffer_length = 0;
// .......... write ..........
inline bool write_compressed_string(const string& out, ostream& os) {
  if (large_buffer_length < out.size() + out.size()/64 + 12) {
    // actually 1.01*out.size() + 12 is enough (according to zlib.h)
    if ( large_buffer != 0 ) delete large_buffer;
    large_buffer_length = out.size()*2 + 12;
    large_buffer = new (unsigned char)[large_buffer_length];
  }

  unsigned long int compressed_size = large_buffer_length;
  int compression_result = compress(large_buffer, &compressed_size,
				    (const unsigned char*)out.c_str(),
				    out.size());
  if ( compression_result != Z_OK ) {
    if ( compression_result == Z_MEM_ERROR )
      global_error()("not enough memory for compression");
    else global_error()("error in compressing");
    return false;
  }
  write_short_int((unsigned short int)out.size(), os); // uncompressed size then
  write_short_int((unsigned short int)compressed_size, os);// compressed size
  os.write(large_buffer, compressed_size);
  return true;
}
// .......... read ..........
inline bool read_compressed_string
(istream& is, pair<unsigned char*, unsigned long int>* result) {

  unsigned short int uncompressed_size = read_short_int(is);
  unsigned short int   compressed_size = read_short_int(is);
  if (small_buffer_length < compressed_size) {
    if ( small_buffer != 0 ) delete small_buffer;
    small_buffer_length = compressed_size * 2;
    small_buffer        = new (unsigned char)[small_buffer_length];
  }
  if ( !is.read(small_buffer, compressed_size) ) return false;

  if (large_buffer_length < uncompressed_size) {
    if ( large_buffer != 0 ) delete large_buffer;
    large_buffer_length = uncompressed_size * 2;
    large_buffer        = new (unsigned char)[large_buffer_length];
  }

  unsigned long int actual_uncompressed_size = uncompressed_size;
  int compression_result = uncompress(large_buffer, &actual_uncompressed_size,
				      small_buffer,           compressed_size);
  if ( compression_result != Z_OK ) {
    if ( compression_result == Z_MEM_ERROR )
      global_error()("not enough memory for uncompression");
    else global_error()("error in uncompressing");
    return false;
  }
  else if ( actual_uncompressed_size != uncompressed_size ) {
    string tmp("error in uncompressing: "
	       "the size of uncompressed data (");
    tmp << actual_uncompressed_size
	<< ") is not equal to the expected value (" << uncompressed_size << ")";
    global_error()(tmp);
    return false;
  }
  *result = make_pair(large_buffer, actual_uncompressed_size);
  return true;
}
// --------------------------------------------------  

void HPT_Setup::write_hits(ostream& os) {
  string out;
  unsigned char last_byte = 0; // last byte always contains some data which
  // should be flushed into the string in the end
  // except for the case when no cells are in the setup and nothing should
  // be written 
  unsigned int ipad = 0;
  for (unsigned int il=0; il<ichambers().size(); ++il) {
    for (unsigned int ip=0; ip<ichambers()[il].size(); ++ip) {
      for (unsigned int ich=0; ich<ichambers()[il][ip].size(); ++ich) {
	for (unsigned int is=0; is<(*ichambers()[il][ip][ich]).second.size();
	     ++is) {
	  HPT_Section& sect = (*ichambers()[il][ip][ich]).second[is];
	  for (HPT_Setup::ICell icell=sect.begin(); icell!=sect.end(); ++icell){
	    if (ipad == 8) {
	      ipad = 0;
	      out += last_byte;
	      last_byte = 0;
	    }
	    if ( (*icell).second.is_fired() == true ) last_byte |= (1<<ipad);
	    ++ipad;
	  }
	}
      }
    }
  }
  if ( out.size() != 0 || ipad != 0 ) out += last_byte;
  
  if ( write_compressed_string(out, os) == false )
    global_error() << "can not write High-Pt compressed event\n";
}

void HPT_Setup::read_hits(istream& is) {
  if ( is.peek() == EOF ) return;
  pair<unsigned char*, unsigned long int> buffer;
  if ( read_compressed_string(is, &buffer) == false ) {
    global_error() << "can not read High-Pt compressed event\n";
    return;
  }
  unsigned int position_in_buffer = 0;
  unsigned char last_byte = buffer.first[0]; // last byte is never
  // counted in position_in_buffer (position_in_buffer should be incremented
  // by one in the end)
  // except for the case when no cells are in the setup and nothing should
  // be read (buffer.second.size == 0)
  unsigned int ipad = 0;
  for (unsigned int il=0; il<ichambers().size(); ++il) {
    for (unsigned int ip=0; ip<ichambers()[il].size(); ++ip) {
      for (unsigned int ich=0; ich<ichambers()[il][ip].size(); ++ich) {
	for (unsigned int isect=0;
	     isect<(*ichambers()[il][ip][ich]).second.size(); ++isect) {
	  HPT_Section& sect = (*ichambers()[il][ip][ich]).second[isect];
	  if ( (buffer.second - position_in_buffer) * 8 < sect.size() ) {
	    string tmp;
	    tmp << "in reading HPT_Setup hits: can not read cell states for "
	      "the section #" << isect << " in layer="
		<< il << ", part=" << ip << ", chamber=" << ich;
	    global_error()(tmp);
	    return;
	  }
	  HPT_Setup::ISection_Cell isc;
	  isc.isection = (*ichambers()[il][ip][ich]).second.begin() + isect;
	  // reset old hits 
	  for (unsigned int i=0; i<sect.fired_cells.size(); ++i)
	    (*sect.fired_cells[i]).second.data = false;
	  sect.fired_cells.clear();
	  // read new ones
	  for (HPT_Setup::ICell icell=sect.begin(); icell!=sect.end(); ++icell){
	    if (ipad == 8) {
	      ipad = 0;
	      ++position_in_buffer;
	      last_byte = buffer.first[position_in_buffer];
	    }
	    if ( (last_byte & (1<<ipad)) != 0 ) {
	      isc.icell = icell;
	      write_data_to_fired_setup_cell(true, isc);
	    }
	    ++ipad;
	  }
	}
      }
    }
  }
  if ( buffer.second != 0 && ++position_in_buffer != buffer.second ) {
    string tmp;
    tmp << "in reading HPT_Setup hits: the total number of bytes given ("
	<< buffer.second
	<< ") is not equal to the number of bytes actually been read ("
	<< position_in_buffer << ")";
    global_error()(tmp);
    return;
  }
}
