/* Usage:

   call_via_istr              (F f, (Arg1*)0, ..., (ArgN*)0); or
   call_via_istr_check_EOF    (F f, (Arg1*)0, ..., (ArgN*)0); or
   call_via_istr_conditionally(F f, (Arg1*)0, ..., (ArgN*)0, Condition c);

	 where F f - some function object which takes 0<=N<=5 arguments and
   expression f(a1, .., aN)
   (a1, .., aN are of types Arg1, .., ArgN respectively)
   is valid.
   If call_*_conditionally(..) is used: Condition c is such an object
   that the expression c(is) is valid ("is" is of type istream*).



   call_via_istr*(...) returns an object of type CallViaIstr0..5<...>
   which has the operator()(istream* is) => it extracts N arguments from "*is",
   if they are extracted successfully ("*is" state is good), "Condition c"
   is applied to "is" and:
   if ( c(is) ) => function f(a1, .., aN) is called.
   In case of simple call_via_istr(..) version condition is substituted by
   the object which always returns true, so that condition has no effect;
	 call_via_istr_check_EOF(..) has a condition, which checks that after
	 extracting of all arguments (*is) is empty.

   The difference in comparison with the call_via_istream.hh file is that
   here the reading errors are sent to the "global_error()" object
   (see "global_error.hh").
   If you are using "global_error()" mechanism, use this file.
   Another difference is the condition mechanism described above.
*/

// Author V.Balagura (balagura@mail.desy.de)   (10.04.98).

#ifndef _call_via_istr_0par_hh
#define _call_via_istr_0par_hh 1

#include <iostream.h>
#include "ptset/global_error.hh"

/* -------------------- 0 args -------------------- */
template <class F, class Condition>
class CallViaIstr0 {
public:
  F function;
  Condition condition;

  CallViaIstr0(F f, Condition c = Condition()) : function(f), condition(c) {}
  void operator()(istream* is) {
    if ( condition(is) ) function();
  }
};
template <class F, class Condition>
inline CallViaIstr0<F, Condition> call_via_istr_conditionally
(F f, Condition c) {
  return CallViaIstr0<F, Condition>(f, c);
}
/* -------------------- 1 arg -------------------- */
template <class F, class A1, class Condition>
class CallViaIstr1 {
public:
  F function;
  Condition condition;

  CallViaIstr1(F f, Condition c = Condition()) : function(f), condition(c) {}
  void operator()(istream* is) {
    A1 a1;
    *is >> a1;
    if ( !(*is) ) { global_error()("can not extract argument"); return; }
    if ( condition(is) ) function(a1);
    else global_error() << a1;
  }
};
template <class F, class A1, class Condition>
inline CallViaIstr1<F, A1, Condition> call_via_istr_conditionally
(F f, A1*, Condition c) {
  return CallViaIstr1<F, A1, Condition>(f, c);
}
/* -------------------- 2 args -------------------- */
template <class F, class A1, class A2, class Condition>
class CallViaIstr2 {
public:
  F function;
  Condition condition;

  CallViaIstr2(F f, Condition c = Condition()) : function(f), condition(c) {}
  void operator()(istream* is) {
    A1 a1;
    A2 a2;
    *is >> a1;
    if ( !(*is) ) { global_error()("can not extract first argument"); return; }
    *is >> a2;
    if ( !(*is) )
      { global_error()("can not extract second argument"); global_error() << a1; return; }
    if ( condition(is) ) function(a1, a2);
    else global_error() << a2 << a1;
  }
};
template <class F, class A1, class A2, class Condition>
inline CallViaIstr2<F, A1, A2, Condition> call_via_istr_conditionally
(F f, A1*, A2*, Condition c) {
  return CallViaIstr2<F, A1, A2, Condition>(f, c);
}
/* -------------------- 3 args -------------------- */
template <class F, class A1, class A2, class A3, class Condition>
class CallViaIstr3 {
public:
  F function;
  Condition condition;

  CallViaIstr3(F f, Condition c = Condition()) : function(f), condition(c) {}
  void operator()(istream* is) {
    A1 a1;
    A2 a2;
    A3 a3;
    *is >> a1;
    if ( !(*is) ) { global_error()("can not extract first argument"); return; }
    *is >> a2;
    if ( !(*is) )
      { global_error()("can not extract second argument"); global_error() << a1; return; }
    *is >> a3;
    if ( !(*is) )
      { global_error()("can not extract third argument"); global_error() << a2 << a1; return; }
    if ( condition(is) ) function(a1, a2, a3);
    else global_error() << a3 << a2 << a1;
  }
};
template <class F, class A1, class A2, class A3, class Condition>
inline CallViaIstr3<F, A1, A2, A3, Condition> call_via_istr_conditionally
(F f, A1*, A2*, A3*, Condition c) {
  return CallViaIstr3<F, A1, A2, A3, Condition>(f, c);
}
/* -------------------- 4 args -------------------- */
template <class F, class A1, class A2, class A3, class A4, class Condition>
class CallViaIstr4 {
public:
  F function;
  Condition condition;

  CallViaIstr4(F f, Condition c = Condition()) : function(f), condition(c) {}
  void operator()(istream* is) {
    A1 a1;
    A2 a2;
    A3 a3;
    A4 a4;
    *is >> a1;
    if ( !(*is) ) { global_error()("can not extract first argument"); return; }
    *is >> a2;
    if ( !(*is) )
      { global_error()("can not extract second argument"); global_error() << a1; return; }
    *is >> a3;
    if ( !(*is) )
      { global_error()("can not extract third argument"); global_error() << a2 << a1; return; }
    *is >> a4;
    if ( !(*is) ) {
      global_error()("can not extract fourth argument"); global_error() << a3 << a2 << a1;
      return;
    }
    if ( condition(is) ) function(a1, a2, a3, a4);
    else global_error() << a4 << a3 << a2 << a1;
  }
};
template <class F, class A1, class A2, class A3, class A4, class Condition>
inline CallViaIstr4<F, A1, A2, A3, A4, Condition> call_via_istr_conditionally
(F f, A1*, A2*, A3*, A4*, Condition c) {
  return CallViaIstr4<F, A1, A2, A3, A4, Condition>(f, c);
}
/* -------------------- 5 args -------------------- */
template <class F, class A1, class A2, class A3, class A4, class A5,
  class Condition>
class CallViaIstr5 {
public:
  F function;
  Condition condition;

  CallViaIstr5(F f, Condition c = Condition()) : function(f), condition(c) {}
  void operator()(istream* is) {
    A1 a1;
    A2 a2;
    A3 a3;
    A4 a4;
    A5 a5;
    *is >> a1;
    if ( !(*is) ) { global_error()("can not extract first argument"); return; }
    *is >> a2;
    if ( !(*is) )
      { global_error()("can not extract second argument"); global_error() << a1; return; }
    *is >> a3;
    if ( !(*is) )
      { global_error()("can not extract third argument"); global_error() << a2 << a1; return; }
    *is >> a4;
    if ( !(*is) ) {
      global_error()("can not extract fourth argument"); global_error() << a3 << a2 << a1;
      return;
    }
    *is >> a5;
    if ( !(*is) ) {
      global_error()("can not extract fifth argument"); global_error() << a4 << a3 << a2 << a1;
      return;
    }
    if ( condition(is) ) function(a1, a2, a3, a4, a5);
    else global_error() << a5 << a4 << a3 << a2 << a1;
  }
};
template <class F, class A1, class A2, class A3, class A4, class A5,
  class Condition>
inline CallViaIstr5<F, A1, A2, A3, A4, A5, Condition> 
call_via_istr_conditionally
(F f, A1*, A2*, A3*, A4*, A5*, Condition c) { 
  return CallViaIstr5<F, A1, A2, A3, A4, A5, Condition>(f, c); 
}

/* -------------------- call_via_istr -------------------- */
/* No condition on istream is required */
struct _IstrOk {
  bool operator()(istream*) { return true; }
};

template <class F>
inline CallViaIstr0<F, _IstrOk> call_via_istr(F f) { 
  return CallViaIstr0<F, _IstrOk>(f, _IstrOk()); 
}
template <class F, class A1>
inline CallViaIstr1<F, A1, _IstrOk> call_via_istr(F f, A1*) { 
  return CallViaIstr1<F, A1, _IstrOk>(f, _IstrOk()); 
}
template <class F, class A1, class A2>
inline CallViaIstr2<F, A1, A2, _IstrOk> call_via_istr(F f, A1*, A2*) {
  return CallViaIstr2<F, A1, A2, _IstrOk>(f, _IstrOk()); 
}
template <class F, class A1, class A2, class A3>
inline CallViaIstr3<F, A1, A2, A3, _IstrOk> 
call_via_istr(F f, A1*, A2*, A3*) { 
  return CallViaIstr3<F, A1, A2, A3, _IstrOk>(f, _IstrOk()); 
}
template <class F, class A1, class A2, class A3, class A4>
inline CallViaIstr4<F, A1, A2, A3, A4, _IstrOk> 
call_via_istr(F f, A1*, A2*, A3*, A4*) { 
  return CallViaIstr4<F, A1, A2, A3, A4, _IstrOk>(f, _IstrOk()); 
}
template <class F, class A1, class A2, class A3, class A4, class A5>
inline CallViaIstr5<F, A1, A2, A3, A4, A5, _IstrOk> 
call_via_istr(F f, A1*, A2*, A3*, A4*, A5*) { 
  return CallViaIstr5<F, A1, A2, A3, A4, A5, _IstrOk>(f, _IstrOk()); 
}

/* ----------------- call_via_istr_check_EOF ----------------- */
/* Condition: istream is empty */
struct _CheckIstr_EOF {
  bool operator()(istream* is) {
    *is >> ws;
    if ( (*is).peek() != EOF ) {
      string s("error: not needed extra argument given: \"");
      char c;
      while ( (*is).get(c) ) s += c;
      s += '\"';
      global_error()(s);
      return false;
    }
    else return true;
  }
};

template <class F>
inline CallViaIstr0<F, _CheckIstr_EOF> call_via_istr_check_EOF(F f) { 
  return CallViaIstr0<F, _CheckIstr_EOF>(f, _CheckIstr_EOF()); 
}
template <class F, class A1>
inline CallViaIstr1<F, A1, _CheckIstr_EOF> call_via_istr_check_EOF(F f, A1*) { 
  return CallViaIstr1<F, A1, _CheckIstr_EOF>(f, _CheckIstr_EOF()); 
}
template <class F, class A1, class A2>
inline CallViaIstr2<F, A1, A2, _CheckIstr_EOF> 
call_via_istr_check_EOF(F f, A1*, A2*) {
  return CallViaIstr2<F, A1, A2, _CheckIstr_EOF>(f, _CheckIstr_EOF()); 
}
template <class F, class A1, class A2, class A3>
inline CallViaIstr3<F, A1, A2, A3, _CheckIstr_EOF> 
call_via_istr_check_EOF(F f, A1*, A2*, A3*) { 
  return CallViaIstr3<F, A1, A2, A3, _CheckIstr_EOF>(f, _CheckIstr_EOF()); 
}
template <class F, class A1, class A2, class A3, class A4>
inline CallViaIstr4<F, A1, A2, A3, A4, _CheckIstr_EOF> 
call_via_istr_check_EOF(F f, A1*, A2*, A3*, A4*) { 
  return CallViaIstr4<F, A1, A2, A3, A4, _CheckIstr_EOF>(f, _CheckIstr_EOF()); 
}
template <class F, class A1, class A2, class A3, class A4, class A5>
inline CallViaIstr5<F, A1, A2, A3, A4, A5, _CheckIstr_EOF> 
call_via_istr_check_EOF(F f, A1*, A2*, A3*, A4*, A5*) { 
  return CallViaIstr5<F, A1, A2, A3, A4, A5, _CheckIstr_EOF>
    (f, _CheckIstr_EOF()); 
}

#endif
