
// *****************************************************************************
//
// source:    rcModSel.C
//
// type:      source code
//
// created:   25 Jul 2000
//	
// author:    Andreas Gellrich
//            HERA-B Collaboration
//            Humboldt-University Berlin
//            Physics Department
//            Invalidenstr. 110
//            10115 Berlin
//            Germany
//            E-mail: Andreas.Gellrich@desy.de
//
// symbol:    rcModuleSelect
//
// called by: rc functions
//
// purpose:   steer calling of reconstruction modules
//
// hints: slt = 0x1   SLTEE
//        slt = 0x2   SLTMM
//        slt = 0x4   SLTDV
//        slt = 0x8   SLTEEB
//        slt = 0x10  SLTCUT
//
// changes:
// 25 Jul 2000 (AG) creation
// 26 Jul 2000 (AG) introduced rcSelectSLT
// 27 Jul 2000 (AG) corrected bug
// 28 Jul 2000 (AG) polished printout
// 02 Aug 2000 (AG) added more criterias
// 03 Aug 2000 (AG) introduced RICH occupancy cut handling
// 04 Aug 2000 (AG) commissioned code
// 05 Aug 2000 (AG) polished printout
// 07 Aug 2000 (AG) set error flag in EVRC cuts were changed
// 08 Aug 2000 (AG) corrected RICH occupancy cut and added seed cut
// 11 Aug 2000 (AG) introduced new module selection mode
// 16 Aug 2000 (AG) used new word in EVRC
//
// *****************************************************************************

// standard includes
#include <iostream.h>
#include <stdlib.h>
#include <string.h>

// ARTE includes
#include "arte/arte.hh"
#include "farm/farm.hh"
#include "rang/zfpar.hh"
#include "rise/RingSearchAttrib.h"
#include "reco/rcmodsel.hh"

// externals
extern int FADEB;

//extern int selStandAlone;
//extern int selVDS;
//extern int selITR;
//extern int selHPT;
//extern int selOTR;
//extern int selRICH;
//extern int selTRD;
//extern int selECAL;
//extern int selECAL2;
//extern int selMUON;
//extern int selMatch;
//extern int selRefit;
//extern int selParticleID;
//extern int selVertex;
//extern int selSelectDecays;
//extern int selClassify;

extern int L4TSKP; // skip event

extern int MODSEL; // module selection mode

// prototypes
int rcSelectSLT();

// constants

// RICH S/A
int looseOccuCut  = 3000; // used for selected
int strictOccuCut = 1000; // used for non-selected

// RICH SEED
int looseSeedCut  = 3000; // used for selected
int strictSeedCut = 0;    // used for non-selected

// OTR
float looseQpmax  = 1.0;  // used for selected
float strictQpmax = 0.2;  // used for non-selected

//
// function code
//
int
rcModSel()
{
  ARTE_mark( "rcModSel 0" , 0 );

  int rc = 0;

  //
  // MODE 1 "SLTBASED" *********************************************************
  //
  if( MODSEL == 1 ){

    static int first = 1;
    if( first ){
      first = 0;

      // get values
      char *richOccuMin = getenv( "RICHOCCUMIN" );
      if( richOccuMin ) looseOccuCut = atoi( richOccuMin );

      char *richOccuMax = getenv( "RICHOCCUMAX" );
      if( richOccuMax ) strictOccuCut = atoi( richOccuMax );

      char *richSeedMin = getenv( "RICHSEEDMIN" );
      if( richSeedMin ) looseSeedCut = atoi( richSeedMin );

      char *richSeedMax = getenv( "RICHSEEDMAX" );
      if( richSeedMax ) strictSeedCut = atoi( richSeedMax );
      
      char *otrPcutMin  = getenv( "OTRPCUTMIN" );
      if( otrPcutMin ) looseQpmax = atof( otrPcutMin );
      
      char *otrPcutMax  = getenv( "OTRPCUTMAX" );
      if( otrPcutMax ) strictQpmax = atof( otrPcutMax );

      // print
      cout << " MODSEL: RICH S/A  "
	   << "looseOccuCut="  << looseOccuCut  << " "
	   << "strictOccuCut=" << strictOccuCut << " "
	   << endl;
      cout << " MODSEL: RICH SEED "
	   << "looseSeedCut="  << looseSeedCut  << " "
	   << "strictSeedCut=" << strictSeedCut << " "
	   << endl;
      cout << " MODSEL: OTR       "
	   << "looseQpmax="  << looseQpmax  << " "
	   << "strictQpmax=" << strictQpmax << " "
	   << endl;
    }

    //
    // default settings ***
    //

    // OTR
    zfpar_.qpmax = strictQpmax; // strict cut
    
    // RICH S/A
    RingSearchAttrib::overwriteOccupancyCut = strictOccuCut; // strict cut

    // RICH SEED
    RingSearchAttrib::seedOccupancyCut = strictSeedCut; // strict cut

    // fill EVRC
    int size = 0;
    if( ( size = ArteTable<EVRC>::size() ) ){
      ArtePointer<EVRC> evrc( size );
      evrc->selmask |= RECO_OTR_PCUT;
      evrc->selmask |= RECO_RICH_OCUT;
      if( FADEB ){
	cout << "MODSEL: EVRC selmask=0x" << evrc->selmask << endl;
      }
    }
    
    //
    // selection ***
    //

    // SLT
    ARTE_mark( "rcSelectSLT" , 0 );
    int slt = rcSelectSLT();
    if( FADEB ) cout << " MODSEL SLTBASED: slt=0x" << hex << slt << dec << endl;
    
    // SLT    
    if( ( slt & 0x4  )                       || // SLTDV
        ( slt & 0x8  )                       || // SLTEEB
        ( ( slt & 0x2  ) && ( slt & 0x10 ) )    // SLTMM && SLTCUT 
	){
      
      // OTR
      zfpar_.qpmax = looseQpmax; // loose cut
      
      // RICH S/A
      RingSearchAttrib::overwriteOccupancyCut = looseOccuCut; // loose cut
      
      // RICH SEED
      RingSearchAttrib::seedOccupancyCut = looseSeedCut; // loose cut
      
      // indicates selection
      rc = 1;
    }
    
    // print
    cout << " MODSEL: SLTBASED "
	 << "SLT=0x" << hex << slt << dec << " "
	 << "CHANGES: "
	 << "OTR:qmax="              << zfpar_.qpmax                            << " "
	 << "RICH:occupancyCut="     << RingSearchAttrib::overwriteOccupancyCut << " "
	 << "RICH:seedOccupancyCut=" << RingSearchAttrib::seedOccupancyCut      << " "
	 << endl;
  }
  
  ARTE_mark( "rcModSel 1" , 0 );
  return rc;
}
