
// *****************************************************************************
// 
//  Author:  Rainer Mankel                     Created:  17-Apr-99
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
//
// rcevnt.C
// reconstruction main program
// dispatch selecting the appropriate main program according to
// time or geometry version
//
// changes:
// 17 Dec 1999 (AG) introduce timing
//  8 Feb 2000 (TO) reconstruction module is selected by flags
// 26 May 2000 (AG) store timing in global variable
//  7 Jun 2000 (AG) initialize timers to 0; fill new table EVTI
//  8 Jun 2000 (AG) use ARTE_mark()
// 14 Jun 2000 (AG) invoke event classification here
// 16 Jun 2000 (AG) introduce rcFillEVTI() 
// 10 Jul 2000 (AG) check for incmomming rpm messages
// 25 Jul 2000 (AG) classified FLT and RND triggers here
// 31 Jul 2000 (AG) added flag = flagdst
// 02 Jul 2000 (AG) polished printout
// 03 Aug 2000 (AG) purged obsolete l4t.hh
// 06 Aug 2000 (AG) all rhp moved to dq
// 11 Aug 2000 (AG) introduced flag selClassify
//
// *****************************************************************************

// standard includes
#include <math.h>

// ONLINE includes

// ARTE includes
#include "arte/arte.hh"
#include "arte/ArteFlag.h"
#include "arte/ArteMessage.hh"
#include "arte/Chrono.hh"
#include "arte/ChronoManager.hh"
#include "farm/farm.hh"
#include "reco/rcflags.hh"
#include "reco/reco.hh"
#include "reco/recoOptions.hh"
#include "reco/eclass.hh"
#include "wax/OTrSelect.h"

// prototypes
extern "C" void rcZevnt_();
unsigned int    acFlag();
int             rcClassify(); // event classification
void            rcFillEVTI(); // fill time table EVTI
int             faCheckRhp(); // check for incomming messages
int             l4tEvnt();    // L4 trigger

// globals
float time_rcevnt     = 0;
float time_rcClassify = 0;

int selClassify = 1;

// externals
extern recoOptions RecoOpt;
extern int         FATIME;
extern int         FADEB;
extern int         *classSum; // event classification

//
// program code
//
extern "C" void
rcevnt_()
{
  // mark
  ARTE_mark( "rcevnt_ 0", 0 );

  //---------------------------------------------------------------------------
  // initialize the reco error flag
  //---------------------------------------------------------------------------
  if( recoFlag ) rcSetError("init");

  //---------------------------------------------------------------------------
  // create new EVRC bank if reco is switched on 
  //---------------------------------------------------------------------------
  ArtePointer<EVRC> evrc;

  if( recoFlag ){
    int size = ArteTable<EVRC>::size();
    mmbook_( ArtePointer<EVRC>::getLink().getRef(), 1 );
    evrc = ArtePointer<EVRC>( size + 1 );
    evrc->ipaddr   = arteflag_.ipaddr;
    evrc->errmask  = 0x0;
    evrc->date     = time( NULL );
    evrc->cputime  = 0.0;
    evrc->alca     = acFlag();
    evrc->artevers = arteflag_.artevers;
  }

  //---------------------------------------------------------------------------
  // start timing
  //---------------------------------------------------------------------------
  static Chrono *cron_rcevnt     = 0;
  static Chrono *cron_rcClassify = 0;

  if( !cron_rcevnt ){
    ChronoManager CM;
    CM.instance();

    cron_rcevnt     = CM.createChrono( "rcevnt" );
    cron_rcClassify = CM.createChrono( "rcClassify" ); 
  }
  cron_rcevnt->start();

  //---------------------------------------------------------------------------
  // event reconstruction
  //---------------------------------------------------------------------------
  if( recoFlag ){
    ARTE_mark( "rc99Aevnt", 0 );
    rc99Aevnt();
  } 
  else {
    if (RecoOpt.CheckActiveAlg("RITER")  || 
	RecoOpt.CheckActiveAlg("CATS")   ||
	RecoOpt.CheckActiveAlg("HOLMES") ||
	RecoOpt.CheckActiveAlg("MARPLE") ||
	RecoOpt.CheckActiveAlg("RIRE")   ||
	RecoOpt.CheckActiveAlg("RITER")  ||
	RecoOpt.CheckActiveAlg("VERTEX"))  rcZevnt_();
  }
  
  //---------------------------------------------------------------------------
  // event classification
  //---------------------------------------------------------------------------
  if( recoFlag && selClassify ){
    ARTE_mark( "rcClassify", 0 );
    cron_rcClassify->start();
    rcClassify();
    cron_rcClassify->stop();
    time_rcClassify = cron_rcClassify->lastTime();
    
    if( FATIME ) cout << " RECO: class=" << time_rcClassify << " sec" << endl;
  } 
  
  //---------------------------------------------------------------------------
  // FCS trigger code
  //---------------------------------------------------------------------------

  // tcode
  int tcode = faDDAQFCS_tcode();

  // EVHD table
  ArtePointer<EVHD> evhd = *( ArteTable<EVHD>::end() - 1 );
  
  // 15 FLT triggers
  if( ( tcode == 0 ) || ( tcode == 8 ) ){
    classSum[15-1]++;
    evhd->flagdst |= ECLASS_FLTTRG; // bit-wise OR
  }
  
  // 16 random triggers
  if( ( tcode == 8 ) || ( tcode == 9 ) ){
    classSum[16-1]++;
    evhd->flagdst |= ECLASS_RNDTRG; // bit-wise OR
  }

  // print
  if( FADEB ){
    cout << " CLASS: flagdst=0x" << hex << evhd->flagdst << dec << endl;
    cout << " CLASS: statistics: ";
    for( int i=0; i<16; i++ ){
      cout << "#" << i+1 << "=" << classSum[i] << " "; 
    }
    cout << endl;
  }

  //---------------------------------------------------------------------------
  // set flag correctly (NEEDED!)
  //---------------------------------------------------------------------------
  evhd->flag = evhd->flagdst;

  //-------------------------------------------------------------------
  // fill timing table EVTI
  //-------------------------------------------------------------------  
  if( recoFlag ) rcFillEVTI();

  //-------------------------------------------------------------------
  // stop timing
  //-------------------------------------------------------------------
  cron_rcevnt->stop();
  time_rcevnt = cron_rcevnt->lastTime();

  //-------------------------------------------------------------------
  // fill in EVRC bank 
  //-------------------------------------------------------------------
  if( recoFlag ) {
    evrc->cputime = time_rcevnt;
    evrc->errmask = recoErrorFlag();
    OTrEventTime( evrc->ottmean,  evrc->ottwidth);
  }

  //-------------------------------------------------------------------
  // L4 trigger   
  //-------------------------------------------------------------------
  ARTE_mark( "l4tEvnt", 0 );
  int l4tflag = l4tEvnt();

  // mark  
  ARTE_mark( "rcevnt_ 1", 0 );
}
