// *****************************************************************************
//
// source:
//
// type:      source code
//
// created:   21. Aug 2000
//
// author:    Thorsten Glebe
//            HERA-B Collaboration
//            Max-Planck-Institut fuer Kernphysik
//            Saupfercheckweg 1
//            69117 Heidelberg
//            Germany
//            E-mail: T.Glebe@mpi-hd.mpg.de
//
// Description: Track class
//
// changes:
// 21 Aug 2000 (TG) creation
// 24 Aug 2000 (TG) include arte.hh header file
// 06 Sep 2000 (TG) changed call to t_V.propagate()
//
// *****************************************************************************
#include <iostream>
#include <iomanip>
#include "arte/arte.hh"
#include "vt++/VtTrack.hh"
#include "vt++/VtKalman.hh"
#include "vt++/VtRelation.hh"

using namespace MATRIX;

namespace VERTEX {

  //============================================================================
  // Track::Constructors
  //============================================================================
  Track::Track():
    t_arteptr(0),
    t_isrtra(false),
    t_Q(0),
    t_rm(0.),
    t_p(VtVector(6)),
    t_V(CMatrix()) {}

  Track::Track(const ArtePointer<RTRA>& rtra):
    t_arteptr(getIDX<RTRA>(rtra)),
    t_isrtra(true),
    t_Q(sgn(rtra->pf)),
    t_rm(0.),
    t_p(MATRIX::VtVector(rtra->xf,rtra->yf,rtra->zf,rtra->txf,rtra->tyf,
			 (rtra->pf!=0)? fabs(1./rtra->pf) : 1.e50)),
    t_V(rtra) {
    propagate(0.);          // propagate track to z=0.
    t_G = t_V.dsinv();
  }
  
  Track::Track(const ArtePointer<RSEG>& rseg):
    t_arteptr(getIDX<RSEG>(rseg)),
    t_isrtra(false),
    t_Q(sgn(rseg->pf)),
    t_rm(0.),
    t_p(MATRIX::VtVector(rseg->xf,rseg->yf,rseg->zf,rseg->txf,rseg->tyf,
			 (rseg->pf!=0)? fabs(1./rseg->pf) : 1.e50)),
    t_V(rseg) {
    propagate(0.);          // propagate track to z=0.
    t_G = t_V.dsinv();
  }
  
  Track::Track(const ArtePointer<MTRA>& mtra):
    t_arteptr(getIDX<MTRA>(mtra)),
    t_isrtra(false),
    t_Q(0),
    t_rm(0.),
    t_p(MATRIX::VtVector(mtra->bver->x,mtra->bver->y,mtra->bver->z,
			 mtra->px/mtra->pz,mtra->py/mtra->pz,
			 fabs(sqrt(mtra->px*mtra->px+mtra->py*mtra->py+
				   mtra->pz*mtra->pz)))),
    t_V(mtra) {
    propagate(0.);          // propagate track to z=0.
    t_G = t_V.dsinv();
  }

  Track::Track(const Track* const track):
    t_arteptr(track->t_arteptr),
    t_isrtra (track->t_isrtra),
    t_Q      (track->t_Q),
    t_rm     (0.),
    t_p      (track->t_p),
    t_V      (track->t_V) {
    propagate(0.);          // propagate track to z=0.
    t_G = t_V.dsinv();
  }

  Track::Track(const Track& rhs):
    t_arteptr(rhs.t_arteptr),
    t_isrtra (rhs.t_isrtra),
    t_Q      (rhs.t_Q),
    t_rm     (0.),
    t_p      (rhs.t_p),
    t_V      (rhs.t_V) {
    propagate(0.);          // propagate track to z=0.
    t_G = t_V.dsinv();
  }

  Track::~Track() {}
  
  //============================================================================
  // Track::print
  //============================================================================
  void Track::print( std::ostream& os ) const {
    // os << "[" << t_arteptr << "," << this << "]";

    os  << " x: "  << x()
	<< " y: "  << y()
	<< " z: "  << z()
	<< " tx: " << tx()
	<< " ty: " << ty()
	<< " p: "  << p() << endl
	<< " Cov Matrix: " << V();
  }

  //============================================================================
  // Track::propagate propagate Track to z 
  //============================================================================
  void Track::propagate(const double zz) {
    if( zz == z() ) return;

    double dz = zz - z();
    t_p[0] += dz * tx();
    t_p[1] += dz * ty();
    t_p[2]  = zz;
    t_V.propagate(dz); // propagate Track covariance matrix
    return;
  }

} // namespace VERTEX
