#ifndef __OTrSelect__H__
#define __OTrSelect__H__

#define OTR_RELAXED_CUT 16
#define OTR_LOOSE_CUT  1
#define OTR_STRICT_CUT 2
#define OTR_NHIT1_CUT  4
#define OTR_NHIT2_CUT  8

//======================================================
// Function to select run without coasting beam in OTR.
// Should be run after waxevnt() function.
//
// Example of code in usevnt.C:
//   #include "wax/OTrSelect.h"
//   waxevnt();
//   if (OTrSelect(OTR_STRICT_CUT, 900, 0)) {
//      printf("This event has been rejected\n");
//      return;
//   }
//   // Some ranger reconstruction can be done here...
//
//  Possible values of flag:
//     OTR_STRICT_CUT - to make strict cut on mean TDC count and its RMS
//     OTR_LOOSE_CUT - to make more loose cut
//     OTR_RELAXED_CUT - to make even looser cut
//
//  For nhit1 (lowest number of OTR hits) one can take values 300 - 1000.
//  For run 14154 best value is around 900.
//  This will throw away the majority of events without interaction.
//  
//  For nhit2 (maximum allowed number of OTR hits)
//  no default value is recommended.
//  For alignment studies nhit2=10000 can be reasonable.
//  Value 0 will mean no rejection on nhit2.
//  
//  A.Lanyov            April 2000
//
//======================================================

extern "C" int OTrEventTime(float& mean, float& width) ;

extern "C" int OTrSelect(int flag, int nhit1, int nhit2);

/********************************************************************
Selection of events without coasting beam.

These functions to select a good event without coasting beam
return 1 if event is good, and 0 otherwise.
 
OTr_cut_loose defines a loose cut, OTr_cut_strict defines a more strict cut.
Parameters are mean TDC count and its RMS calculated by all 5-mm OTR cells.

In run 14154 (5 MHz) loose cut selects 27% of all events, strict - 20%.
These numbers are so small since these cuts also throw away empty events:
If one looks only events with nhit>300, the percentages will be 46% and 34%,
and if nhit>1000, then - 76% and 58%, respectively.

In addition a cut nhit>300 can be useful.
Here nhit is total number of TDC counts in the range [16-250] in 5-mm
OTR cells.

*********************************************************************/

/**
  Returns 1 if point (mean,rms) lies inside the good region
  and 0 otherwise.

  The region is specified by a number of points.
 */
extern "C" int OTr_cut_relaxed(float mean, float rms);

/**
  Returns 1 if point (mean,rms) lies inside the good region
  and 0 otherwise.

  The region is specified by rectangular borders.
 */
extern "C" int OTr_cut_loose(float mean, float rms);

/**
  Returns 1 if point (mean,rms) lies inside the good region
  and 0 otherwise.

  The region is specified by rectangular borders.
 */
extern "C" int OTr_cut_strict(float mean, float rms);

#endif
